var   WIP_VERSION = "3.0.9";
var   WIP_UniqPrefix = "WorldIP";

var   WIPCc = Components.classes;
var   WIPCi = Components.interfaces;

var   WIP_STATE_START = WIPCi.nsIWebProgressListener.STATE_START;
var   WIP_STATE_TRANSFERRING = WIPCi.nsIWebProgressListener.STATE_TRANSFERRING;
var   WIP_STATE_STOP = WIPCi.nsIWebProgressListener.STATE_STOP;
var   WIP_STATE_IS_NETWORK = WIPCi.nsIWebProgressListener.STATE_IS_NETWORK;
var   WIP_STATE_IS_DOCUMENT = WIPCi.nsIWebProgressListener.STATE_IS_DOCUMENT;

var   WIP_TYPE_ONE_SHOT = WIPCi.nsITimer.TYPE_ONE_SHOT;
var   WIP_TYPE_REPEATING_SLACK = WIPCi.nsITimer.TYPE_REPEATING_SLACK;

var   WIP_PLATFORM = WIPCc["@mozilla.org/xre/app-info;1"].getService(WIPCi.nsIXULAppInfo).platformVersion;
var   WIP_OS = WIPCc["@mozilla.org/xre/app-info;1"].getService(WIPCi.nsIXULRuntime).OS;

var   WIP_ERROR_MODULE_NETWORK = 2152398848;
var   WIP_ERROR_CONNECTION_REFUSED = WIP_ERROR_MODULE_NETWORK + 13;
var   WIP_ERROR_NET_TIMEOUT = WIP_ERROR_MODULE_NETWORK + 14;
var   WIP_ERROR_NET_RESET = WIP_ERROR_MODULE_NETWORK + 20;
var   WIP_DNS_CODE_ERROR_FORMAT = 1;
var   WIP_DNS_CODE_ERROR_SERVFAIL = 2;
var   WIP_DNS_CODE_ERROR_NAME = 3;
var   WIP_DNS_CODE_ERROR_NOT_IMPLEMENTED = 4;
var   WIP_DNS_CODE_ERROR_REFUSED = 5;

var WIP_HEX = ["0","1","2","3","4","5","6","7","8","9","a","b","c","d","e","f"];
var WIP_DNS_QTYPE = {"A":"01", "NS":"02", "SOA":"06", "MX":"0f", "TXT":"10", "PTR":"0c", "AAAA":"1c", "ANY":"ff", "CNAME":"05", "SPF":"63", "NAPTR":"23", "SRV":"21", "RRSIG":"2e"};
var WIP_DNS_RQTYPE = {"01":"A", "02":"NS", "06":"SOA", "0f":"MX", "10":"TXT", "0c":"PTR", "1c":"AAAA", "05":"CNAME", "63":"SPF", "23":"NAPTR", "21":"SRV", "2e":"RRSIG"};

window.addEventListener("load", function() { wip.init(); }, false);
window.addEventListener("unload", function() { wip.killer(); }, false);

var wipcookieremover =
{
	observe : function(subject, topic, data)
	{
		if (topic != "http-on-modify-request")
			return;
		var channel = subject.QueryInterface( WIPCi.nsIHttpChannel );
		if ( /api[123456]?\.wipmania\.com/.test( channel.originalURI.host ) ) {
			channel.setRequestHeader("Cookie", "", false);
		}
	}
}

var wip_storage =
{
	_cache: {},
	_timeout: {},
	_fuel: false,
	_UniqPrefix: "WIP",

	get: function(key)
	{
		key = String(key);
		if (key in this._cache) {
			if (this._timeout[key] && this.current_seconds() >= this._timeout[key] ) {
					delete this._cache[key];
					delete this._timeout[key];
			} else
				return this._cache[key];
		}
		if ( this._fuel )
			return this.get_global(key);
		return null;
	},

	get_global: function(key)
	{
		key = String(key);
		if ( this._fuel && this._fuel.storage.has(this._UniqPrefix+key) ) {
			var r = this._fuel.storage.get(this._UniqPrefix+key, null)
			var t = this._fuel.storage.get(this._UniqPrefix+"#t#"+key, 0);
			if (t && this.current_seconds() >= t) {
				this.delete(key);
				return null;
			}
			if ( !( key in this._cache && this._cache[key] == r ) ) {
				this._cache[key] = r;
				this._timeout[key] = t;
			}
			return r;
		}
		return null;
	},

	set: function(key, value, timeout)
	{
		key = String(key);
		if (typeof(timeout)!="number" || !timeout)
			timeout = 0; //forever
		else
			timeout = timeout + this.current_seconds();
		if ( this._fuel ) {
			this._fuel.storage.set(this._UniqPrefix+key, value);
			if (timeout)
				this._fuel.storage.set(this._UniqPrefix+"#t#"+key, timeout);
			else if (this._fuel.storage.has(this._UniqPrefix+"#t#"+key))
				this._fuel.storage.set(this._UniqPrefix+"#t#"+key, null)
		}
		this._cache[key] = value;
		this._timeout[key] = timeout;
	},

	delete: function(key)
	{
		key = String(key);
		if (key in this._cache) {
			delete this._cache[key];
			delete this._timeout[key];
		}
		if ( this._fuel && this._fuel.storage.has(this._UniqPrefix+key) ) {
			this._fuel.storage.set(this._UniqPrefix+key,null)
			if ( this._fuel.storage.has(this._UniqPrefix+"#t#"+key) )
				this._fuel.storage.set(this._UniqPrefix+"#t#"+key, null)
		}
	},

	current_seconds: function()
	{
		var lastTime = new Date().getTime();
		return Math.round(lastTime/1000);
	},
}

var wip =
{
	JSON: JSON,
	prefs: {},
	WIPTimer: {},

	init: function()
	{
		var content = document.getElementById("content");
		wip.lock = 0;
		wip.currentUri = "";
		wip.longLastRequests = 0;
		wip.longLastRequestsDate = 0;
		wip.currentIP = "";
		wip.backupAPI = "";
		wip.dnstcp = "tcp";
		wip.proxy_dns_enabled = false;
		wip.thread = WIPCc["@mozilla.org/thread-manager;1"].getService(WIPCi.nsIThreadManager);
		wip.dns = WIPCc["@mozilla.org/network/dns-service;1"].getService(WIPCi.nsIDNSService);
		wip.tldService = WIPCc['@mozilla.org/network/effective-tld-service;1'].getService(WIPCi.nsIEffectiveTLDService);

		// cache between firefox-windows
		try {
			var WIP_BrowserName = document.getElementById("bundle_brand").getString("brandShortName");
			if ( WIP_BrowserName == "SeaMonkey" )
				wip_storage._fuel = WIPCc["@mozilla.org/smile/application;1"].getService(WIPCi.smileIApplication);
			else if ( WIP_BrowserName == "Mozilla" || WIP_BrowserName == "Firefox" )
				wip_storage._fuel = WIPCc["@mozilla.org/fuel/application;1"].getService(WIPCi.fuelIApplication);
		} catch(e) {}

		wip.worldip_myip_leer = document.getElementById("worldip-myip").getAttribute("label");
		wip.worldip_rdns_leer = document.getElementById("worldip-getrdns").getAttribute("label");
		wip.worldip_ip_leer = document.getElementById("worldip-ip").getAttribute("label");
		var prefService = WIPCc["@mozilla.org/preferences-service;1"];
		wip.preferences = prefService.getService(WIPCi.nsIPrefService).getBranch("extensions.worldip.");
		wip.preferences.QueryInterface(WIPCi.nsIPrefBranch);
		wip.preferencesProxy = prefService.getService(WIPCi.nsIPrefService).getBranch("network.proxy.");
		wip.preferencesProxy.QueryInterface(WIPCi.nsIPrefBranch);
		wip.currentAPI = "api";
		wip.observerService = WIPCc["@mozilla.org/observer-service;1"].getService(WIPCi.nsIObserverService);
		wip.initPrefs();
		wip.cookieDisable();
		wip.prefs["version"] = wip.preferences.getCharPref("version");
		if ( (!wip.prefs["version"] || (wip.prefs["version"] != WIP_VERSION)) && ( wip_storage.get("update_version_lock") != true ) ) {
			wip_storage.set("update_version_lock", true);
			wip.updateVersion();
		}
		wip.updateMenuItems();
		wip.generateMenu();
		wip.preferences.addObserver("", this, false);
		wip.preferencesProxy.addObserver("", this, false);
		wip.observerService.addObserver(this, "http-on-examine-response", false);

		wip.setPositon();

		wip.WIPTimerGarbageCollector = WIPCc["@mozilla.org/timer;1"].createInstance(WIPCi.nsITimer);
		var eventWIPTimerGarbageCollector = { notify: function(timer) { wip.clearTimer(); } };
		try { wip.WIPTimerGarbageCollector.initWithCallback(eventWIPTimerGarbageCollector, 120000, WIP_TYPE_REPEATING_SLACK); } catch(e) {}

		wip.WIPTimerRestoreAPI = WIPCc["@mozilla.org/timer;1"].createInstance(WIPCi.nsITimer);
		var eventWIPTimerRestoreAPI = { notify: function(timer) { wip.restoreMainAPI(); } };
		try { wip.WIPTimerRestoreAPI.initWithCallback(eventWIPTimerRestoreAPI, 12000, WIP_TYPE_REPEATING_SLACK); } catch(e) {}

		wip.WIPTimerActualInfo = WIPCc["@mozilla.org/timer;1"].createInstance(WIPCi.nsITimer);
		var eventWIPTimerActualInfo = { notify: function(timer) {
			var hostActual = wip.getHostActual();
			var ip1 = wip.ipResolve(hostActual, true);
			var ip2 = wip.getCurrentIP();
			if ((!ip1 && hostActual) || (ip1 && ip1!=ip2)) {
				wip.showInfoActual();
				if (!ip1)
					wip.ipResolve(hostActual);
			}
		} };
		try { wip.WIPTimerActualInfo.initWithCallback(eventWIPTimerActualInfo, 1000, WIP_TYPE_REPEATING_SLACK); } catch(e) {}

		wip.MyIPTimer = WIPCc["@mozilla.org/timer;1"].createInstance(WIPCi.nsITimer);
		if (wip.prefs["enablemyip"])
			wip_storage.get_global("myip");
		else if (wip.prefs["enablelands"])
			wip.getMyIP();
		else
			wip.setPanelAttributes();
		wip.showInfo(null,null);

		this.Listener =
		{
			QueryInterface: function(aIID)
			{
				if (aIID.equals(WIPCi.nsIWebProgressListener) ||
				aIID.equals(WIPCi.nsISupportsWeakReference) ||
				aIID.equals(WIPCi.nsIXULBrowserWindow) ||
				aIID.equals(WIPCi.nsISupports))
				return this;
				throw Components.results.NS_NOINTERFACE;
			},
			onLocationChange:function(aProgress, aRequest, aURI)
			{
				try {
					wip.onTabLocationChange(aURI);
				} catch(e) {}
			},
			onProgressChange:function(a,b,c,d,e,f){},
			onSecurityChange:function(a,b,c){},
			onStateChange:function(aProgress,aRequest,aFlag,aStatus) {
				if (aFlag & WIP_STATE_START) {
					try {
						var scheme = aRequest.URI.scheme;
						if ((scheme == "http")||(scheme == "https")||(scheme == "ftp")) {
							if (aRequest && aRequest.URI && aRequest.URI.asciiHost && aRequest.URI.asciiHost==wip.getHostActual())
								wip.ipResolve(aRequest.URI.asciiHost);
						}
					} catch(e) {}
				} else if ((aFlag & WIP_STATE_STOP) && (aFlag & WIP_STATE_IS_NETWORK)) {
				}
			},
			onStatusChange:function(a,b,c,d){},
			onLinkIconAvailable:function(a){}
		};

		try {
			content.addProgressListener(this.Listener);
			document.getElementById("home-button").addEventListener("click", wip.onTabSelect, false);
			window.getBrowser().tabContainer.addEventListener("TabOpen", wip.onTabOpen, false);
			window.getBrowser().addEventListener("select", wip.onTabSelect, true);
		} catch(e) {}
		// self check for spoofing
		if (!wip.ipResolve("api.wipmania.com", true) && (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"]) && (wip.prefs["enablefulldns"] || wip.prefs["enablecheckextdns"]) )
			wip.setTimer(function() { wip.remoteDNSRequest("api.wipmania.com", "A", wip.randomDNSServer(), wip.dnstcp, 0); }, 30, WIP_TYPE_ONE_SHOT);
	},

	killer: function()
	{
		var content = document.getElementById("content");
		content.removeProgressListener(this.Listener);
		window.getBrowser().tabContainer.removeEventListener("TabOpen", wip.onTabOpen, false);
		window.getBrowser().removeEventListener("select", wip.onTabSelect, true);
		wip.preferences.removeObserver("", this);
		wip.preferencesProxy.removeObserver("", this);
		wip.cookieDisableOff();
	},

	clearTimer: function() {
		var nowTime = new Date().getTime()+Math.random();
		var keys = Object.keys(wip.WIPTimer);
		for (k in keys) {
			var t = keys[k];
			if (nowTime-59000>t) {
				delete wip.WIPTimer[t];
			}
		}
	},
	
	setTimer: function(callbackFunc, delay, typ) // for small timers <59sec
	{
		var startTime = new Date().getTime()+Math.random();
		wip.WIPTimer[startTime] = WIPCc["@mozilla.org/timer;1"].createInstance(WIPCi.nsITimer);
		var event = {
			notify: function(timer)
			{
				try {
					callbackFunc.call();
				} catch(e) {}
			}
		};
		try {
			wip.WIPTimer[startTime].initWithCallback(event,delay,typ);
		} catch(e) {}
	},

	initPrefs: function()
	{
		wip.prefs["myiplogenable"] = false;
		wip.proxy_dns_enabled = false;
		wip.prefs["proxy_dns_uselocal"] = wip.preferences.getBoolPref("proxy_dns_uselocal");
		wip.prefs["proxy_type"]= wip.preferencesProxy.getIntPref("type");
		wip.prefs["proxy_socks"]= wip.preferencesProxy.getCharPref("socks");
		wip.prefs["proxy_socks_port"]= wip.preferencesProxy.getIntPref("socks_port");
		wip.prefs["proxy_socks_version"]= wip.preferencesProxy.getIntPref("socks_version");
		wip.prefs["proxy_socks_remote_dns"]= wip.preferencesProxy.getBoolPref("socks_remote_dns");
		if ( wip.prefs["proxy_type"] == 1 )
			wip.proxy_dns_enabled = (wip.prefs["proxy_socks"]!="") && (wip.prefs["proxy_socks_port"]>0) && (wip.prefs["proxy_socks_version"]==5) && wip.prefs["proxy_socks_remote_dns"];
		wip.prefs["enablemyip"]= wip.preferences.getBoolPref("enablemyip");
		wip.prefs["updatemyiptimer"]= wip.preferences.getCharPref("updatemyiptimer");
		if (wip.prefs["updatemyiptimer"]=="m") // minutes
			wip.prefs["updatemyipinterval"]= 60 * wip.preferences.getIntPref("updatemyipinterval");
		else // hours
			wip.prefs["updatemyipinterval"]= 3600 * wip.preferences.getIntPref("updatemyipinterval");
		wip.prefs["position"]= wip.preferences.getCharPref("position"); //s,a,t - statusbar,addressbar,toolbar
		wip.prefs["toolbarposition"]= wip.preferences.getCharPref("toolbarposition"); //left,center,right
		wip.prefs["clickaction"]= wip.preferences.getCharPref("clickaction"); //no,land,copy,newip,wrongland,getland,rdns,dns,asonhe,getall
		wip.prefs["myiplogenable"]= wip.preferences.getBoolPref("myiplogenable");
		wip.prefs["showflag"]= wip.preferences.getBoolPref("showflag");
		wip.prefs["flagset"]= wip.preferences.getCharPref("flagset");
		wip.prefs["showland"]= wip.preferences.getBoolPref("showland");
		wip.prefs["showlandcode"]= wip.preferences.getBoolPref("showlandcode");
		wip.prefs["showcity"]= wip.preferences.getBoolPref("showcity");
		wip.prefs["showcitycolor"]= wip.preferences.getCharPref("showcitycolor");
		wip.prefs["showip"]= wip.preferences.getBoolPref("showip");
		wip.prefs["showmyflag"]= wip.preferences.getBoolPref("showmyflag");
		wip.prefs["myflagset"]= wip.preferences.getCharPref("myflagset");
		wip.prefs["showmyland"]= wip.preferences.getBoolPref("showmyland");
		wip.prefs["showmylandcode"]= wip.preferences.getBoolPref("showmylandcode");
		wip.prefs["showmyip"]= wip.preferences.getBoolPref("showmyip");
		wip.prefs["showlandcolor"]= wip.preferences.getCharPref("showlandcolor");
		wip.prefs["showlandcodecolor"]= wip.preferences.getCharPref("showlandcodecolor");
		wip.prefs["showmylandcolor"]= wip.preferences.getCharPref("showmylandcolor");
		wip.prefs["showmylandcodecolor"]= wip.preferences.getCharPref("showmylandcodecolor");
		wip.prefs["showipcolor"]= wip.preferences.getCharPref("showipcolor");
		wip.prefs["showmyipcolor"]= wip.preferences.getCharPref("showmyipcolor");
		wip.prefs["enablelands"]= wip.preferences.getBoolPref("enablelands");
		wip.prefs["showdatacenter"]= wip.preferences.getBoolPref("showdatacenter");
		wip.prefs["showdatacentercolor"]= wip.preferences.getCharPref("showdatacentercolor");
		wip.prefs["showasnumber"]= wip.preferences.getBoolPref("showasnumber");
		wip.prefs["showasnumbercolor"]= wip.preferences.getCharPref("showasnumbercolor");
		wip.prefs["showasname"]= wip.preferences.getBoolPref("showasname");
		wip.prefs["showasnamecolor"]= wip.preferences.getCharPref("showasnamecolor");
		wip.prefs["showmyasname"]= wip.preferences.getBoolPref("showmyasname");
		wip.prefs["showmyasnamecolor"]= wip.preferences.getCharPref("showmyasnamecolor");
		wip.prefs["showmyasnumber"]= wip.preferences.getBoolPref("showmyasnumber");
		wip.prefs["showmyasnumbercolor"]= wip.preferences.getCharPref("showmyasnumbercolor");
		wip.prefs["showmydatacenter"]= wip.preferences.getBoolPref("showmydatacenter");
		wip.prefs["showmydatacentercolor"]= wip.preferences.getCharPref("showmydatacentercolor");
		wip.prefs["enablerdns"]= wip.preferences.getBoolPref("enablerdns");
		wip.prefs["showrdns"]= wip.preferences.getBoolPref("showrdns");
		wip.prefs["showrdnscolor"]= wip.preferences.getCharPref("showrdnscolor");
		wip.prefs["enablemyrdns"]= wip.preferences.getBoolPref("enablemyrdns");
		wip.prefs["showmyrdns"]= wip.preferences.getBoolPref("showmyrdns");
		wip.prefs["showmyrdnscolor"]= wip.preferences.getCharPref("showmyrdnscolor");
		wip.prefs["acceptinrank"]= wip.preferences.getBoolPref("acceptinrank");
		wip.prefs["enablefulldns"]= wip.preferences.getBoolPref("enablefulldns");
		wip.prefs["enablecheckextdns"]= wip.preferences.getBoolPref("enablecheckextdns");
		wip.prefs["enablespoofpopup"]= wip.preferences.getBoolPref("enablespoofpopup");
		wip.prefs["extdnsservers"]= wip.preferences.getCharPref("extdnsservers");
	},

	updateVersion: function()
	{
		wip.preferences.setCharPref("version",WIP_VERSION);
		var lastTime = new Date().getTime();
		var lastTimeDays = Math.round(lastTime/(24*3600*1000)); //days from Epoch
		wip.preferences.setIntPref("pointstime", lastTimeDays-5); // make first check only tomorrow or later
		wip.setTimer(function() { wip.goHome("apps/"); }, 1000, WIP_TYPE_ONE_SHOT);
		if (wip.prefs["version"] && (wip.prefs["version"].indexOf("1.")==0 || wip.prefs["version"].indexOf("2.")==0) ) {
			try { wip.preferences.deleteBranch("enablehttpserver"); } catch(e) {}
			try { wip.preferences.deleteBranch("apiResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("api1ResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("api2ResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("api3ResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("api4ResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("api5ResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("api6ResponseTime"); } catch(e) {}
			try { wip.preferences.deleteBranch("APIServerListMD5"); } catch(e) {}
			try { wip.preferences.deleteBranch("APIServerFolge"); } catch(e) {}
			try { wip.preferences.deleteBranch("lastmyland"); } catch(e) {}
			try { wip.preferences.deleteBranch("enableflagcache"); } catch(e) {}
		}
		if (wip.prefs["version"] && wip.prefs["version"].indexOf("2.")==0) {
			if (document.getElementById("status-bar").clientHeight) {
				if (wip.prefs["position"]=="a") {
					wip.preferences.setCharPref("position","s");
					wip.prefs["position"]=="s";
				}
			} else {
				if (wip.prefs["position"]=="s") {
					wip.preferences.setCharPref("position","a");
					wip.prefs["position"]=="a";
				}
			}
		}
	},

	updateMenuItems: function()
	{
		var currentTime = new Date().getTime();
		var currentTimeDays = Math.round(currentTime/(24*3600*1000));
		var lastTimeDays = wip.preferences.getIntPref("pointstime");
		wip.prefs["points"]= wip.JSON.parse(wip.preferences.getCharPref("points"));
		if ((currentTimeDays-lastTimeDays)>6) // check one time per week for new list of ping-servers
			wip.setTimer(function() { wip.requestAPI(wip.currentAPI,"points.txt", wip.getNewPointsCallback, {}); }, 1000, WIP_TYPE_ONE_SHOT);
	},

	getNewPointsCallback: function(doc, args)
	{
		var lastTime = new Date().getTime();
		var lastTimeDays = Math.round(lastTime/(24*3600*1000)); //days from Epoch
		wip.preferences.setIntPref("pointstime", lastTimeDays);
		try {
			var newpoints = wip.JSON.parse(doc);
			if (!wip.prefs["points"].v || newpoints.v!=wip.prefs["points"].v){
				wip.preferences.setCharPref("points", doc);
				if (newpoints.m)
					wip.showMessage(newpoints.m,newpoints.mclick);
				wip.updateMenuItems();
				wip.generateMenu();
			}
		} catch(e) {}
	},

	generateSubMenu: function(act,command)
	{
		var points = wip.prefs["points"];
		var stringbundle = document.getElementById("worldip-tooltiplabels");
		var menupopup = document.createElement("menupopup");
		var menuitemworld = document.createElement("menuitem");
			menuitemworld.setAttribute("oncommand", "wip.goHome"+act+"("+command+",'world');");
			menuitemworld.setAttribute("label", stringbundle.getString("worldwide"));
			menuitemworld.setAttribute("class", "menuitem-iconic worldip-menuitem-iconic");
			menuitemworld.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/--.png);");
		menupopup.appendChild(menuitemworld);
		if (points.p) {
			for (point in points.p){
			if (points.p[point].subcontinent) {
						menupopup.appendChild(document.createElement("menuseparator"));
						var kontinent = document.createElement("menuitem");
						try { var kontinentname = stringbundle.getString(points.p[point].subcontinent); } catch(e) { var kontinentname = points.p[point].subcontinent; }
						kontinent.style.fontWeight = "bold";
						kontinent.setAttribute("label", kontinentname);
						menupopup.appendChild(kontinent);
			} else {
				var menuitemcity = document.createElement("menuitem");
					menuitemcity.setAttribute("oncommand", "wip.goHome"+act+"("+command+", '"+point+"');");
					try { var cityname = stringbundle.getString(point); } catch(e) { var cityname = point; }
					menuitemcity.setAttribute("label", cityname+", "+stringbundle.getString(points.p[point]));
					menuitemcity.setAttribute("class", "menuitem-iconic worldip-menuitem-iconic");
					menuitemcity.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/"+points.p[point]+".png);");
				menupopup.appendChild(menuitemcity);
			}
			}
		}
		return menupopup;
	},

	generateMenu: function()
	{
		var ping_ip_menu = document.getElementById("worldip-ping-ip-menu");
		var trace_ip_menu = document.getElementById("worldip-trace-ip-menu");
		var ping_myip_menu = document.getElementById("worldip-ping-myip-menu");
		var trace_myip_menu = document.getElementById("worldip-trace-myip-menu");
		var provider_byname = document.getElementById("worldip-provider-byname");
		var provider_bycity = document.getElementById("worldip-provider-bycity");
		var stringbundle = document.getElementById("worldip-tooltiplabels");
		var points = wip.prefs["points"];

		while (ping_ip_menu.lastChild)
			ping_ip_menu.removeChild(ping_ip_menu.lastChild);
		while (trace_ip_menu.lastChild)
			trace_ip_menu.removeChild(trace_ip_menu.lastChild);
		while (ping_myip_menu.lastChild)
			ping_myip_menu.removeChild(ping_myip_menu.lastChild);
		while (trace_myip_menu.lastChild)
			trace_myip_menu.removeChild(trace_myip_menu.lastChild);
		while (provider_byname.lastChild)
			provider_byname.removeChild(provider_byname.lastChild);
		while (provider_bycity.lastChild)
			provider_bycity.removeChild(provider_bycity.lastChild);

		var submenu = wip.generateSubMenu("Ping","wip.getCurrentIP()");
		ping_ip_menu.appendChild(submenu);
		var submenu = wip.generateSubMenu("Trace","wip.getCurrentIP()");
		trace_ip_menu.appendChild(submenu);
		var submenu = wip.generateSubMenu("Ping","wip.myip()");
		ping_myip_menu.appendChild(submenu);
		var submenu = wip.generateSubMenu("Trace","wip.myip()");
		trace_myip_menu.appendChild(submenu);

		var menuitemproviderbynamepopup = document.createElement("menupopup");
			if (points.pr) {
				for (prov in points.pr){
					var provitem = document.createElement("menuitem");
						provitem.setAttribute("label", points.pr[prov].c);
						provitem.setAttribute("class", "menuitem-iconic worldip-menuitem-iconic");
						provitem.setAttribute("style", "list-style-image: url(chrome://worldip/skin/prov/"+prov+".png);");
						if (points.pr[prov].la) // qwest hack
							provitem.setAttribute("oncommand", "wip.goProvider('"+points.pr[prov].l+"');wip.goProvider('"+points.pr[prov].la+"');");
						else
							provitem.setAttribute("oncommand", "wip.goProvider('"+points.pr[prov].l+"');");
					menuitemproviderbynamepopup.appendChild(provitem);
				}
			}
		provider_byname.appendChild(menuitemproviderbynamepopup);

		var menuitemproviderbycitypopup = document.createElement("menupopup");
			if (points.prc) {
				for (land in points.prc){
				if (points.prc[land].subcontinent) {
						menuitemproviderbycitypopup.appendChild(document.createElement("menuseparator"));
						var kontinent = document.createElement("menuitem");
						try { var kontinentname = stringbundle.getString(points.prc[land].subcontinent); } catch(e) { var kontinentname = points.prc[land].subcontinent; }
						kontinent.style.fontWeight = "bold";
						kontinent.setAttribute("label", kontinentname);
						menuitemproviderbycitypopup.appendChild(kontinent);
				} else {
					var landmenu = document.createElement("menu");
						landmenu.setAttribute("label", stringbundle.getString(land));
						landmenu.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/"+land+".png);");
						landmenu.setAttribute("class", "menu-iconic");
						var landpopup = document.createElement("menupopup");
						for (city in points.prc[land]) {
							var citymenu = document.createElement("menu");
								try { var cityname = stringbundle.getString(city); } catch(e) { var cityname = city; }
								citymenu.setAttribute("label", cityname);
								var citymenupopup = document.createElement("menupopup");
								for (prov in points.prc[land][city]) {
									var menuitemprov = document.createElement("menuitem");
										var url = points.pr[prov].l;
										if (points.pr[prov].la && land!="US") // qwest hack
											url = points.pr[prov].la;
										menuitemprov.setAttribute("oncommand", "wip.goProvider('"+url+"');");
										menuitemprov.setAttribute("label", points.pr[prov].c);
										menuitemprov.setAttribute("class", "menuitem-iconic worldip-menuitem-iconic");
										menuitemprov.setAttribute("style", "list-style-image: url(chrome://worldip/skin/prov/"+prov+".png);");
									citymenupopup.appendChild(menuitemprov);
								}
								citymenu.appendChild(citymenupopup);
							landpopup.appendChild(citymenu);
						}
						landmenu.appendChild(landpopup);
					menuitemproviderbycitypopup.appendChild(landmenu);
				}
				}
			}
		provider_bycity.appendChild(menuitemproviderbycitypopup);
	},

	cookieDisable: function()
	{
		wip.observerService.addObserver(wipcookieremover, "http-on-modify-request", false);
	},

	cookieDisableOff: function()
	{
		wip.observerService.removeObserver(wipcookieremover, "http-on-modify-request");
	},

	observe: function(subject, topic, data)
	{
		if (topic == "http-on-examine-response" && subject.URI && subject.URI.asciiHost) {
			try {
				var host = subject.URI.asciiHost;
				var server = subject.getResponseHeader("Server");
				wip_storage.set("server#"+host, server);
			} catch(e) {}
			return;
		}
		if (topic != "nsPref:changed")
			return;
		if (data=="type" || data=="socks" || data=="socks_port" || data=="socks_version" || data=="socks_remote_dns" || data=="proxy_dns_uselocal")
		{
			wip.initPrefs();
			wip.changeMyIP();
			return;
		}
		if (data=="enablemyip")
			wip.enableMyIP(0);
		else if (data=="updatemyipinterval" || data=="updatemyiptimer")
		{
			wip.initPrefs();
			wip.changeMyIP();
		}
		else if (data=="flagset")
		{
			wip.prefs["flagset"] = wip.preferences.getCharPref("flagset");
			var panel_flag = document.getElementById("worldip-statusbar-flag");
			var src = panel_flag.getAttribute("src");
			if (wip.prefs["flagset"]=="0")
				src = src.replace("flags1","flags0");
			else
				src = src.replace("flags0","flags1");
			panel_flag.setAttribute("src", src);
			wip.setPanelAttributes();
			wip.generateMenu();
		}
		else if (data=="myflagset")
		{
			wip.prefs["myflagset"] = wip.preferences.getCharPref("myflagset");
			var panel_flag = document.getElementById("worldip-statusbar-myflag");
			var src = panel_flag.getAttribute("src");
			if (wip.prefs["myflagset"]=="0")
				src = src.replace("flags1","flags0");
			else
				src = src.replace("flags0","flags1");
			panel_flag.setAttribute("src", src);
			wip.setPanelAttributes();
			wip.generateMenu();
		}
		else if (data=="myiplogenable") {
			wip_storage.delete("myiplog");
			wip.initPrefs();
			wip.setPanelAttributes();
		}
		else
		{
			wip.initPrefs();
			wip.setPositon();
			wip.showInfoActual();
		}
	},

	onTabOpen: function(e)
	{
		var b = window.getBrowser().getBrowserForTab(e.target);
		wip.isHostReady(e.target,0);
	},

	isHostReady: function(target,time)
	{
		if (time > 4000)
			return false;
		var b = window.getBrowser().getBrowserForTab(target);
		try
		{
			var host = b.currentURI.host;
			var scheme = b.currentURI.scheme;
			if (!((scheme == "http")||(scheme == "https")||(scheme == "ftp")))
				return false;
			wip.ipResolve(host);
		}
		catch(e)
		{
			wip.setTimer(function() { wip.isHostReady(target,time+300); }, 300, WIP_TYPE_ONE_SHOT);
		}
	},

	onTabLocationChange: function(Location)
	{
		if (!wip.lock)
		{
			if (Location && (Location.scheme != "chrome") && (Location.scheme != "file") && (Location.scheme != "about") && Location.host )
			{
				var ip = wip.ipResolve(Location.host);
				wip.currentUri = Location;
				wip.currentIP = ip;
				if (!ip)
					wip.showInfo(null,Location.host);
				else
					wip.showInfo(ip,Location.host);
			}
			else
			{
				try
				{
					var host = Location.host;
				} catch(e)
				{
					var host = null;
					if ( Location.scheme == "chrome" || Location.scheme == "file" )
						host = Location;
				}
				wip.showInfo(null,host);
				wip.currentUri = "";
				wip.currentIP = "";
			}
		}
		wip.lock = 0;
		return true;
	},

	onTabSelect: function(e)
	{
		var nodeName = e.originalTarget.nodeName;
		if (nodeName == "xul:tabs" || nodeName == "toolbarbutton")
		{
			wip.currentUri = "";
			wip.currentIP = "";
			if (nodeName == "toolbarbutton")
				wip.showInfo(null,null);
			return false;
		}
		var selectedTab = e.originalTarget.selectedIndex;
		if (!e.originalTarget.childNodes || !e.originalTarget.childNodes[selectedTab]) {
			wip.currentUri = "";
			wip.currentIP = "";
			wip.showInfo(null,null);
			return;
		}
		var selectedDoc = e.originalTarget.childNodes[selectedTab].firstChild;
		if (!selectedDoc || !selectedDoc.contentDocument) {
			wip.currentUri = "";
			wip.currentIP = "";
			wip.showInfo(null,null);
			return;
		}
		var uri = selectedDoc.contentDocument.location;
		wip.lock=1;
		try {
			var ip = wip.ipResolve(uri.host);
			if (ip) {
				wip.currentUri = uri;
				wip.currentIP = ip;
				wip.showInfo(ip,uri.host);
			} else {
				wip.currentUri = uri;
				wip.currentIP = "";
				wip.showInfo("",uri.host);
			}
		} catch(e)
		{
			wip.currentUri = "";
			wip.currentIP = "";
			wip.showInfo(null,null);
		}
		return true;
	},

	ipRemoteResolveCallback: function(doc, args)
	{
		var pattern = /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/;
		if (doc && !pattern.test(doc)) doc = 0;
		if (!doc && doc==="") doc = "";
		wip_storage.set("dns#"+args.host, [doc,]);
		if (doc) {
			if (wip.prefs["enablecheckextdns"]) {
				wip_storage.set("spoof#"+args.host, 4, 3600);
				wip_storage.set("spooftext#"+args.host, "SpoofText_socks", 3600);
			}
			wip.getData(doc,args.host,true);
		}
	},

	ipRemoteResolve: function()
	{
		if (!wip.currentUri && !wip.currentUri.host)
			return null;
		var host = wip.currentUri.host.split(':')[0];
		if (/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(host))
			return null;
		var d = wip_storage.get("dns#"+host);
		if (d && d[0]) {
			if (d[0].indexOf(':')>0) {
				d.sort(function(a,b){return a.indexOf(":") >= b.indexOf(":")});
				return d[0];
			}
			else
				return d[0];
		}
		else if (d != null && (d === 0 || (d.length && (d[0] === 0 || d[0] === "")) ) )
			return null;
		wip.setTimer(function() { wip.requestAPI(wip.currentAPI, "dns/?d="+host, wip.ipRemoteResolveCallback, {"host":host,}); }, 50, WIP_TYPE_ONE_SHOT);
	},

	removeDuplicate: function(a)
	{
		var _temp = {};
		var temp = [];
		for(var i=0; i<a.length; i++){
			if (!(a[i] in _temp)) {
				_temp[a[i]] = true;
				temp.push(a[i]);
			}
		}
		return temp;
	},


	dnsAsyncRequest: function(host, force, forceData, doubleCheck)
	{
		if (!host || (/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(host)) ) {
			wip.loaderDisable(host);
			return;
		}
		var dnsCallback = {
			onLookupComplete: function(aRequest, resolved, aStatus) {
				if (aStatus || !resolved || !resolved.hasMore()) // failed
				{
					wip_storage.set("dns#"+host, 0, 3600);
					return;
				}
				var ips = [];
				var ip = "";
				try {
					ip = resolved.getNextAddrAsString();
					ips.push(ip);
				} catch(e) {}
				if (ip && resolved.hasMore()) {
					try {
						while (resolved.hasMore()) {
							var ip2 = resolved.getNextAddrAsString();
							ips.push(ip2);
						}
					} catch(e) {}
				}
				if (ips.length) {
					ips = wip.removeDuplicate(ips);
					var sorted_ips = ips;
					sorted_ips.sort(function(a,b){return a.indexOf(":") >= b.indexOf(":")});
					ip = sorted_ips[0];
					ips = sorted_ips;
				} else {
					ip = "";
				}
				if (doubleCheck) {
					var d = wip_storage.get("dns#"+host);
					if (d && d[0] && ip==d[0]) return null;
				}
				if ( (wip.prefs["enablecheckextdns"] && !wip_storage.get("comparearecslock#"+host) && (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"]) && (!doubleCheck || (doubleCheck && ip) ) ) || forceData ) {
					wip_storage.set("comparearecslock#"+host, true, 1);
					wip.setTimer(function() { wip.compareArecords(host, 1); }, 500, WIP_TYPE_ONE_SHOT);
				}
				var pattern = /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/;
				if (pattern.test(ip) && ip.indexOf(":")==-1)
				{
					wip_storage.set("dns#"+host, ips, 3600);
					wip.getData(ip,host,(forceData ? true : false));
					if ( (wip.prefs["enablerdns"] && (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"]) ) || forceData)
						wip.getRDNS(ip);
				}
				else if ( ip == "::1" )
				{
					wip_storage.set("dns#"+host, ["127.0.0.1", "::1"], 3600);
					wip.showInfoActual();
				}
				else // we doesn't work with IPv6
				{
					wip_storage.set("dns#"+host, ips, 3600);
					wip.showInfoActual(); // "V6"
				}
			}
		};

		try {
			wip.dns.asyncResolve(host, (force ? 1 : 0), dnsCallback, wip.thread.mainThread);
		} catch(ex) {
			wip_storage.set("dns#"+host, 0, 3600);
		}

		if ((!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"] || forceData) && !doubleCheck) {
			if (wip.prefs["enablefulldns"] || forceData)
				wip.setTimer(function() { wip.remoteDNS(host, "ANY"); }, 30, WIP_TYPE_ONE_SHOT);
			else if (wip.prefs["enablecheckextdns"]) {
				wip.setTimer(function() { wip.remoteDNS(host, "A"); }, 30, WIP_TYPE_ONE_SHOT);
				//wip.remoteDNS(host, "AAAA");
				var tophost1 = host.split(".").slice(1).join(".");
				if (tophost1!=wip.tldService.getPublicSuffixFromHost(tophost1))
					wip.setTimer(function() { wip.remoteDNS(tophost1, "A"); }, 50, WIP_TYPE_ONE_SHOT);
				var tophost2 = wip.getTophost(host);
				if (tophost2 && tophost1!=tophost2)
					wip.setTimer(function() { wip.remoteDNS(tophost2, "A"); }, 70, WIP_TYPE_ONE_SHOT);
					//wip.remoteDNS(tophost, "AAAA");
			}
		}
	},


	ipResolve: function(fullhost, onlyCache) // on resolv return always IPv4, but in the sequence we have really list
	{
		if (!fullhost) return null;
		if (/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(fullhost)) {
			if (!wip_storage.get("dns#"+fullhost))
				wip_storage.set("dns#"+fullhost, [fullhost,], 24*3600);
			if (!onlyCache) {
				wip.getData(fullhost,fullhost,false);
				if (wip.prefs["enablerdns"] && (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"]) )
					wip.getRDNS(fullhost);
			}
			return fullhost;
		}

		var host = fullhost.split(':')[0];
		if (host.length == 4) {
			if (host.indexOf('.')<0)
				host = fullhost; // IPv6
		}
		var d = wip_storage.get("dns#"+host);
		if (d && d[0]) {
			if (!onlyCache && !wip_storage.get("dnslock#"+host)) {
				wip_storage.set("dnslock#"+host, 1, 120);
				wip.dnsAsyncRequest(host, false, false, true);
			}
			if (d[0].indexOf(':')>0) {
				d.sort(function(a,b){return a.indexOf(":") >= b.indexOf(":")});
				return d[0];
			}
			else
				return d[0];
		}
		else if (d != null && (d === 0 || (d.length && (d[0] === 0 || d[0] === "")) ) )
			return null;

		if (wip.proxy_dns_enabled && !wip.prefs["proxy_dns_uselocal"])
			return null;

		if (host.indexOf(':')>0){
			wip_storage.set("dns#"+host, [host], 3600);
			return host;
		}
		if (!onlyCache)
			wip.dnsAsyncRequest(host, false, false, false);
		return null;
	},

	isPublicNetz: function(ip)
	{
		if (!ip)
			return false;
		var pattern = /^(0\.|10\.|127\.|192\.168|169\.254|198\.18\.|198\.19\.|192\.0\.0\.|192\.0\.2\.|198\.51\.100\.|203\.0\.113\.|192\.88\.99|2[3-5]\d|22[4-9]).*/; //RFC 3330,1918,2544,3171,1700,5736
		if (pattern.test(ip))
			return false;
		else if ((ip.indexOf("172.") == 0) && (16 <= parseInt(ip.split(".")[1])) && (parseInt(ip.split(".")[1]) <= 31)) // RFC 1918
			return false;
		else if ((ip.indexOf("100.") == 0) && (64 <= parseInt(ip.split(".")[1])) && (parseInt(ip.split(".")[1]) <= 127)) // RFC 6598
			return false;
		return true;
	},

	getIANAStatus: function(ip)
	{
		if (!ip)
			return false;
		return wipIPspace[ip.split(".")[0]];
	},

	getPrivateType: function(ip)
	{
		if (!ip)
			return false;
		if (ip.indexOf("127.") == 0)
			return "localhost";
		else if ((ip.indexOf("10.") == 0) || (ip.indexOf("192.168.") == 0))
			return "intnetwork";
		else if ((ip.indexOf("172.") == 0) && (16 <= parseInt(ip.split(".")[1])) && (parseInt(ip.split(".")[1]) <= 31))
			return "intnetwork";
		else if ((ip.indexOf("100.") == 0) && (64 <= parseInt(ip.split(".")[1])) && (parseInt(ip.split(".")[1]) <= 127))
			return "RFC6598";
		else if (ip.indexOf("0.") == 0)
			return "thisnetwork";
		else if (ip.indexOf("169.254.") == 0)
			return "linklocal";
		else if ((ip.indexOf("192.0.0.") == 0) || (ip.indexOf("192.0.2.") == 0) || (ip.indexOf("198.51.100.") == 0) || (ip.indexOf("203.0.113.") == 0))
			return "testnet";
		else if ((ip.indexOf("198.18.") == 0) || (ip.indexOf("198.19.") == 0))
			return "networktesting";
		else if (ip.indexOf("192.88.99") == 0)
			return "relay6to4";
		else if (ip == "V6" || ip.indexOf(":") >= 0)
			return "V6";
		else
			return false;
	},

	loaderShow: function(host) {
		var l = wip_storage.get("loader#"+host);
		if (l) {
			var statusbar_loader = document.getElementById("worldip-statusbar-loader");
			statusbar_loader.setAttribute("hidden", "false");
			var addressbar_loader = document.getElementById("worldip-addressbar-loader");
			addressbar_loader.setAttribute("hidden", "false");
		} else {
			var statusbar_loader = document.getElementById("worldip-statusbar-loader");
			statusbar_loader.setAttribute("hidden", "true");
			var addressbar_loader = document.getElementById("worldip-addressbar-loader");
			addressbar_loader.setAttribute("hidden", "true");
		}
	},

	loaderEnable: function(host) {
		wip_storage.set("loader#"+host, true, 30);
		wip.loaderShow(host);
	},

	loaderDisable: function(host) {
		wip_storage.delete("loader#"+host);
		wip.loaderShow(host);
	},

	getAndCheckAll: function()
	{
		if (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"]) { //local dns resolver
			if (!wip.currentUri && !wip.currentUri.host)
				return null;
			var host = wip.currentUri.host.split(':')[0];
			wip.loaderEnable(host);
			wip.setTimer(function() { wip.loaderDisable(host); }, 15000, WIP_TYPE_ONE_SHOT);
			wip.dnsAsyncRequest(host, true, true, false);
		}
		else { // external dns resolver via https
			wip.ipRemoteResolve();
		}
	},

	getDisabledLand: function()
	{
		var ip = wip.getCurrentIP();
		var host = wip.currentUri.host;
		wip.prefs["enablelands"] = true;
		wip.getData(ip,host, true);
		wip.prefs["enablelands"] = false;
	},

	callbackRestoreMainAPI: function(doc, args)
	{
		wip.backupAPI = "";
		wip.currentAPI = "api";
		wip_storage.delete("restoremainip");
	},

	restoreMainAPI: function()
	{
		if (wip.currentAPI=="api") return;
		wip.setTimer(function() { wip.requestAPI("api","",wip.callbackRestoreMainAPI, {}); }, 50, WIP_TYPE_ONE_SHOT);
	},

	changeCurrentAPI: function()
	{
		if (!wip.backupAPI && wip.currentAPI=="api") {
			//DE,US,JP,CL,AU,SG = 123456
			var ipcache = wip_storage.get("ip"+wip.myip());
			var myland = (ipcache && ipcache.l) ? ipcache.l : "";
			if (myland in ["US","CA","MX","CR","BZ","GT","PA","HN","NI","SV","AG","AN","BB","BS","DM","GD","HT","KN","LC","MQ","PR","TT","VG","AI","AW","BM","CU","DO","GP","JM","KY","MF","MS","TC","VC","VI","GU"])
				wip.backupAPI = "214365";
			if (myland in ["AR","BR","CO","FK","GY","PY","UY","BO","CL","EC","GF","PE","SR","VE"])
				wip.backupAPI = "421365";
			if (myland in ["AT","CH","FR","LU","NL","BE","DE","LI","MC","AX","EE","FO","GG","IM","JE","LV","SE","DK","FI","GB","IE","IS","LT","NO","AD","BA","GI","HR","ME","MT","RS","SM","AL","ES","GR","IT","MK","PT","SI","VA","BG","CZ","MD","RO","SK","BY","HU","PL","RU","UA","AE","AZ","CY","IL","JO","LB","PS","SA","TR","AM","BH","GE","IQ","KW","OM","QA","SY","YE","KG","TJ","UZ","KZ","TM"])
				wip.backupAPI = "126345";
			if (myland in ["JP","KR","TW","CN","HK","MO","MN"])
				wip.backupAPI = "362514";
			if (myland in ["AU","NZ"])
				wip.backupAPI = "526341";
			if (myland in ["AF","BT","IO","LK","NP","BD","IN","IR","MV","PK","BN","KH","MM","PH","SG","TL","ID","LA","MY","PW","TH","VN"])
				wip.backupAPI = "635214";
			else
				wip.backupAPI = "123645";
		}
		if (wip.backupAPI) {
			wip.currentAPI = "api"+wip.backupAPI[0];
			wip.backupAPI = wip.backupAPI.substring(1);
		} else { // oops, we have tried all servers. wait...
			wip.currentAPI = "";
		}
	},

	requestAPI: function(currentAPI,args,callbackFunc,ccallbackArgs)
	{
		if (!currentAPI || wip_storage.get("iplock"+currentAPI+args)) return;
		if (!ccallbackArgs["httpstries"])
			ccallbackArgs["httpstries"] = 1;
		else
			ccallbackArgs["httpstries"] += 1;
		var reqTimeoutSec = Math.round(ccallbackArgs["httpstries"]/2)+3;
		wip_storage.set("iplock"+currentAPI+args, true, reqTimeoutSec);
		var req = new XMLHttpRequest();
		var scheme = (args && args!="points.txt") ? "https" : "http";
		if (ccallbackArgs && ccallbackArgs["viahttp"]) scheme = "http";
		req.open("GET", scheme+"://"+currentAPI+".wipmania.com/"+args, true);
		req.setRequestHeader("Accept", "text/html,text/plain");
		req.setRequestHeader("Accept-Encoding", null);
		//req.setRequestHeader("Keep-Alive", null);
		req.setRequestHeader("Accept-Charset", null);
		//req.setRequestHeader("Connection", "close");
		if (wip.prefs["acceptinrank"])
			req.setRequestHeader("Parameter", "rank");
		req.setRequestHeader("Typ", "ffplugin2");
		req.setRequestHeader("Ver", WIP_VERSION);
		req.timeout = reqTimeoutSec*1000;
		req.ontimeout = function () {
			wip_storage.set("iplock"+currentAPI+args, false);
			if ((ccallbackArgs["host"] || ccallbackArgs["ip"]) && !wip_storage.get("iplocktimeout"+currentAPI+args)) {
				if (ccallbackArgs["viahttp"]) delete ccallbackArgs["viahttp"];
				else ccallbackArgs["viahttp"] = 1;
				if (wip.currentAPI==currentAPI && !ccallbackArgs["viahttp"])
					wip.changeCurrentAPI();
				if (ccallbackArgs["httpstries"]<6)
					wip.setTimer(function() { wip.requestAPI(wip.currentAPI,args,callbackFunc,ccallbackArgs); }, 40, WIP_TYPE_ONE_SHOT);
				else
					wip.loaderDisable(ccallbackArgs["host"]);
			}
		};
		req.onreadystatechange = function()
		{
			try {
				if (req.readyState === 4)
				{
					wip_storage.set("iplock"+currentAPI+args, false);
					if (req.status === 200) {
						callbackFunc(req.responseText, ccallbackArgs);
					} else if (req.status === 0) {
						wip_storage.set("iplock"+currentAPI+args, false);
						wip_storage.set("iplocktimeout"+currentAPI+args, 1);
						if (ccallbackArgs["host"] || ccallbackArgs["ip"]) {
							if (ccallbackArgs["viahttp"]) delete ccallbackArgs["viahttp"];
							else ccallbackArgs["viahttp"] = 1;
							if (wip.currentAPI==currentAPI && !ccallbackArgs["viahttp"])
								wip.changeCurrentAPI();
							if (ccallbackArgs["httpstries"]<6)
								wip.setTimer(function() { wip.requestAPI(wip.currentAPI,args,callbackFunc,ccallbackArgs); }, 40, WIP_TYPE_ONE_SHOT);
							else
								wip.loaderDisable(ccallbackArgs["host"]);
						}
					}
				}
			} catch (e) {}
		};
		req.send(null);
	},

	dataCallback: function(doc,args)
	{
		try {
			var apiresults = wip.JSON.parse(doc);
		} catch(e) {
			var apiresults = {"a":"","n":"","d":"","r":"","l":"--","c":""};
		}
		if (apiresults && apiresults.l && args && args.ip)
		{
			if (!apiresults.r || apiresults.r==0) apiresults.r = "";
			if (!apiresults.r && args && args.ip) {
				var ipcache = wip_storage.get("ip"+args.ip);
				if (ipcache && ipcache.r) apiresults.r = ipcache.r;
			}
			if (!apiresults.a || apiresults.a==0) apiresults.a = "";
			if (!apiresults.n || apiresults.n==0) apiresults.n = "";
			if (!apiresults.c || apiresults.c==0) apiresults.c = "";
			if (!apiresults.d || apiresults.d==0) apiresults.d = "";
			if (apiresults.l=="XX" || apiresults.l=="AP" || apiresults.l=="EU" || apiresults.l=="--") {
				if (apiresults.l=="XX" && apiresults.c) {
					apiresults.l = "--";
					wip_storage.set("ip"+args.ip, {"l":apiresults.l,"a":apiresults.a,"r":apiresults.r,"c":apiresults.c}, 24*3600); // Land, AS number, RDNS, City and Region code
				} else
					wip_storage.set("ip"+args.ip, {"l":apiresults.l,"a":apiresults.a,"r":apiresults.r,"c":apiresults.c}, 3600);
			}
			else
				wip_storage.set("ip"+args.ip, {"l":apiresults.l,"a":apiresults.a,"r":apiresults.r,"c":apiresults.c}, 24*3600);
			if (!wip_storage.get("as"+apiresults.a))
				wip_storage.set("as"+apiresults.a, {"n":apiresults.n,"d":apiresults.d}); // AS name, Provider/Datacenter
			wip.showInfoActual();
		}
		if (args.host) {
			var spoof = wip_storage.get("spoof#"+args.host);
			if (spoof && spoof>2) {
				wip.setTimer(function() { wip.compareArecords(args.host, 6); }, 100, WIP_TYPE_ONE_SHOT);
				wip.setTimer(function() { wip.loaderDisable(args.host); }, 10000, WIP_TYPE_ONE_SHOT);
			} else
				wip.loaderDisable(args.host);
		}
	},

	getData: function(ip,host,force)
	{
		var ipcache = wip_storage.get("ip"+ip);
		if (wip.isPublicNetz(ip) && ip!="V6" && ip.indexOf(":")<0 && (wip.prefs["enablelands"] || !host || (!wip.prefs["enablelands"] && force) ))
		{
			if (ipcache && ipcache.l && (host || ipcache.a) && !force ) {
				//wip.showInfo(ip,host);
				wip.showInfoActual();
			}
			else {
				var args = ip ? (host ? ip+"?"+host : ip) : "";
				if (host)
					wip.loaderEnable(host);
				wip.setTimer(function() { wip.requestAPI(wip.currentAPI, args, wip.dataCallback, {"ip":ip,"host":host,}); }, 50, WIP_TYPE_ONE_SHOT);
			}
		}
		else
		{
			if (!wip.prefs["enablelands"])
				if (!ipcache)
					wip_storage.set("ip"+ip, {"l":"--","a":"","r":"","c":""});
			//wip.showInfo(ip,host);
			wip.showInfoActual();
		}
	},

	myip: function()
	{
		var myip = wip_storage.get("myip");
		if (myip || myip==="")
			return myip;
		var lastmyip = wip_storage.get("myiplast");
		wip.setTimer(function() { wip.changeMyIP(); }, 50, WIP_TYPE_ONE_SHOT);
		if (lastmyip)
			return lastmyip;
		return "";
	},

	getMyIPcallback: function(doc, args)
	{
		try
		{
			var myData = doc.split("<br>");
			var myip = myData[0].substring(0,15);
			var myland = myData[1].substring(0,2);
			var myiplast = wip_storage.get("myiplast");
			if (myip && wip.prefs["myiplogenable"] && (!myiplast || myiplast != myip)) {
				var myiplog = wip_storage.get("myiplog");
				var zeit = new Date().toLocaleString();
				if (!myiplog)
					myiplog = new Array();
				if (myiplog.length==0 || myiplog[myiplog.length-1].ip != myip) {
					myiplog.push({"ip":myip,"time":String(zeit)});
					wip_storage.set("myiplog", myiplog);
				}
			}
			if (wip.prefs["updatemyipinterval"] > 0)
				wip_storage.set("myip", myip, wip.prefs["updatemyipinterval"]);
			else
				wip_storage.set("myip", myip);
			wip_storage.set("myiplast", myip);
			if (wip.prefs["enablemyrdns"])
				wip.getRDNS(myip);
			var ipcache = wip_storage.get("ip"+myip);
			if (ipcache == null || ipcache.l != myland) {
				wip_storage.set("ip"+myip, {"l":myland,"a":"","r":"","c":""});
				wip.getData(myip,"");
			} else if (!wip_storage.get("as"+ipcache.a)) {
				wip.getData(myip,"");
			}
		}
		catch(e)
		{
			if (wip.prefs["updatemyipinterval"] > 0)
				wip_storage.set("myip", "", wip.prefs["updatemyipinterval"]);
			else
				wip_storage.set("myip", "");
		}
		wip.setPanelAttributes();
	},

	getMyIP: function()
	{
		wip.setTimer(function() { wip.requestAPI(wip.currentAPI,"",wip.getMyIPcallback, {"host":"-",}); }, 30, WIP_TYPE_ONE_SHOT);
	},

	getHostActual: function() {
		var br = getBrowser();
		if (!br || !br.contentDocument)
			return null;
		var Location = br.contentDocument.location;
		var host = null;
		if (Location && Location.host && (Location.scheme != "chrome") && (Location.scheme != "file") && (Location.scheme != "about"))
			host = Location.host;
		else if ( Location.scheme == "chrome" || Location.scheme == "file" )
			host = Location;
		if (!host) return null;
		var host2 = host.split(':')[0];
		if (host2.length == 4) {
			if (host2.indexOf('.')<0)
				host2 = host; // IPv6
		}
		return host2;
	},

	showInfoActual: function()
	{
		var host = wip.getHostActual();
		var ip = wip.ipResolve(host, true);
		ip = ip ? ip : null;
		wip.currentUri = getBrowser().contentDocument.location;
		wip.currentIP = ip;
		wip.showInfo(ip,host);
		wip.loaderShow(host);
	},

	showInfo: function(ip,host)
	{
		var ipcache = wip_storage.get("ip"+ip);
		if (ipcache && ipcache.l && wip.isPublicNetz(ip)) {
			var flagimage = ipcache.l;
			var country = ipcache.l;
			if (country=="--") {
				if (ipcache.c && ipcache.c=="000")
					country = "";
				else if (wip.prefs["enablelands"])
					country = "";
			}
		}
		else
		{
			var country = "";
			var flagimage = "--";
			if (ip && (ip=="V6" || ip.indexOf(":")>=0))
				country = "V6";
			else if (ip && !wip.isPublicNetz(ip))
			{
				if (host)
				{
					var PrivateType = wip.getPrivateType(ip);
					var flagimage = (PrivateType=="localhost") ? "20" : "21";
					country = PrivateType ? PrivateType : "localhost";
				}
			}
			else if (ip && wip.currentAPI)
			{
				wip.setTimer(function() { wip.showInfoActual(); }, 300, WIP_TYPE_ONE_SHOT);
				wip.getData(ip,host);
				return;
			}
			if (!ip && host && wip.proxy_dns_enabled && !PrivateType) {
				country = "--";
				flagimage = "--";
			}
		}
		if (!wip.prefs["showflag"])
			flagimage = "--";

		var stringbundle = document.getElementById("worldip-tooltiplabels");
		var panel = document.getElementById("worldip-statusbar");
		var panel_flag = document.getElementById("worldip-statusbar-flag");
		var panel_land = document.getElementById("worldip-statusbar-land");
		var panel_landcode = document.getElementById("worldip-statusbar-landcode");
		var panel_city = document.getElementById("worldip-statusbar-city");
		var panel_ip = document.getElementById("worldip-statusbar-ip");
		panel_flag.setAttribute("src", "chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/"+flagimage+".png");
		panel_landcode.setAttribute("value", (ipcache && ipcache.l) ? ipcache.l : "");
		if (country) {
			if (country=="--" && wip.proxy_dns_enabled)
				panel_land.setAttribute("value", "DNS "+stringbundle.getString(country)+" (SOCKSv5)");
			else
				panel_land.setAttribute("value", stringbundle.getString(country));
		} else
			panel_land.setAttribute("value", "");
		if (ipcache && ipcache.c)
			panel_city.setAttribute("value", ipcache.c);
		else
			panel_city.setAttribute("value", "");

		var showip = ip ? ip : "";
		var d = wip_storage.get("dns#"+host);
		if (d && d.length>1)
			showip = showip + " (+" + (d.length-1) + ")"
		panel_ip.setAttribute("value", showip);
		wip.setPanelAttributes();
		return true;
	},

	setPositon: function()
	{
		switch (wip.prefs["position"])
		{
			case "s": //statusbar
				try
				{	// this function is originally from flagfox extension. Rewrited for compatibility with add-on "Statusable"
					var statusbar = document.getElementById("status-bar");
					var worldipbar = document.getElementById("worldip-statusbar");
					switch (wip.prefs["toolbarposition"])
					{
						case "left":
							var el = statusbar.firstChild;
							break;
						case "center":
							var el = document.getElementById("statusbar-progresspanel").nextSibling;
							if (!el)
								el = document.getElementById("download-monitor").nextSibling;
							break;
						case "right":
							var el = statusbar.lastChild.nextSibling;
					}
					statusbar.insertBefore(worldipbar,el);
				} catch(e) {}
				break;
			case "a": //addressbar
				try
				{
					var worldipbar = document.getElementById("worldip-addressbar");
					switch (wip.prefs["toolbarposition"])
					{
						case "left":
							var addressbar = document.getElementById("urlbar");
							var el = addressbar.firstChild.nextSibling;
							break;
						case "center":
							var addressbar = document.getElementById("urlbar-icons");
							var el = addressbar.firstChild;
							break;
						case "right":
							var addressbar = document.getElementById("urlbar-icons");
							var el = addressbar.lastChild.nextSibling;
					}
					addressbar.insertBefore(worldipbar, el);
				} catch(e) {}
				break;
		}
	},

	setPanelAttributesStatusbar: function(ip)
	{
		var panel_warning_red = document.getElementById("worldip-statusbar-warning-r");
		var panel_warning_yellow = document.getElementById("worldip-statusbar-warning-y");
		var panel_warning_green = document.getElementById("worldip-statusbar-warning-g");
		var panel_land = document.getElementById("worldip-statusbar-land");
		var panel_landcode = document.getElementById("worldip-statusbar-landcode");
		var panel_city = document.getElementById("worldip-statusbar-city");
		var panel_ip = document.getElementById("worldip-statusbar-ip");
		var panel_datacenter = document.getElementById("worldip-statusbar-datacenter");
		var panel_asnumber = document.getElementById("worldip-statusbar-asnumber");
		var panel_asname = document.getElementById("worldip-statusbar-asname");
		var panel_rdns = document.getElementById("worldip-statusbar-rdns");
		var panel_myflag = document.getElementById("worldip-statusbar-myflag");
		var panel_myland = document.getElementById("worldip-statusbar-myland");
		var panel_mylandcode = document.getElementById("worldip-statusbar-mylandcode");
		var panel_myip = document.getElementById("worldip-statusbar-myip");
		var panel_mydatacenter = document.getElementById("worldip-statusbar-mydatacenter");
		var panel_myasnumber = document.getElementById("worldip-statusbar-myasnumber");
		var panel_myasname = document.getElementById("worldip-statusbar-myasname");
		var panel_myrdns = document.getElementById("worldip-statusbar-myrdns");
		var stringbundle = document.getElementById("worldip-tooltiplabels");

		var hostActual = wip.getHostActual();
		var spoof = 0;
		if (hostActual)
			var spoof = wip_storage.get("spoof#"+wip.getHostActual());
		panel_warning_green.setAttribute("hidden", (spoof === 1 || spoof === 2) ? "false" : "true");
		panel_warning_yellow.setAttribute("hidden", (spoof === 3) ? "false" : "true");
		panel_warning_red.setAttribute("hidden", (spoof === 4) ? "false" : "true");
		
		panel_landcode.setAttribute("hidden", wip.prefs["showlandcode"] ? "false" : "true");
		if (panel_landcode.getAttribute("value") == "")
			panel_landcode.setAttribute("hidden", "true");
		panel_landcode.setAttribute("style", "color:"+wip.prefs["showlandcodecolor"]+";");
		
		panel_land.setAttribute("hidden", wip.prefs["showland"] ? "false" : "true");
		if (panel_land.getAttribute("value") == "")
			panel_land.setAttribute("hidden", "true");
		panel_land.setAttribute("style", "color:"+wip.prefs["showlandcolor"]+";");

		panel_ip.setAttribute("hidden", wip.prefs["showip"] ? "false" : "true");
		if (panel_ip.getAttribute("value") == "")
			panel_ip.setAttribute("hidden", "true");
		panel_ip.setAttribute("style", "color:"+wip.prefs["showipcolor"]+";");

		var myip = wip.myip();
		panel_myip.setAttribute("value", myip ? myip : "");
		panel_myip.setAttribute("style", "color:"+wip.prefs["showmyipcolor"]+";");
		panel_myip.setAttribute("hidden", (wip.prefs["showmyip"] && myip && wip.prefs["enablemyip"]) ? "false" : "true");

		var myipcache = wip_storage.get("ip"+myip);
		var myascache = "";
		var myflagland = "--";
		if (myipcache && myipcache.a)
			myascache = wip_storage.get("as"+myipcache.a);
		if (myipcache && myipcache.l)
			myflagland = myipcache.l;
		
		panel_myflag.setAttribute("src", "chrome://worldip/skin/flags"+wip.prefs["myflagset"]+"/"+myflagland+".png");
		panel_myflag.setAttribute("hidden", wip.prefs["showmyflag"] ? "false" : "true");

		panel_mylandcode.setAttribute("hidden", wip.prefs["showmylandcode"] ? "false" : "true");
		panel_mylandcode.setAttribute("value", myflagland)
		if (panel_mylandcode.getAttribute("value") == "")
			panel_mylandcode.setAttribute("hidden", "true");
		panel_mylandcode.setAttribute("style", "color:"+wip.prefs["showmylandcodecolor"]+";");
		
		panel_myland.setAttribute("hidden", wip.prefs["showmyland"] ? "false" : "true");
		try {
				panel_myland.setAttribute("value", stringbundle.getString(myflagland));
		} catch(e) {}
		if (panel_myland.getAttribute("value") == "")
			panel_myland.setAttribute("hidden", "true");
		panel_myland.setAttribute("style", "color:"+wip.prefs["showmylandcolor"]+";");

		if (myascache) {
			var mydatacenter = myascache.d;
			if (mydatacenter.length>40)
				mydatacenter = mydatacenter.substring(0,40)+"…";
			panel_mydatacenter.setAttribute("value", mydatacenter);
			panel_mydatacenter.setAttribute("style", "color:"+wip.prefs["showmydatacentercolor"]+";");
			panel_mydatacenter.setAttribute("hidden", (myascache.d && wip.prefs["showmydatacenter"] && myip && wip.prefs["enablemyip"]) ? "false" : "true");

			panel_myasnumber.setAttribute("value", "AS"+myipcache.a);
			panel_myasnumber.setAttribute("style", "color:"+wip.prefs["showmyasnumbercolor"]+";");
			panel_myasnumber.setAttribute("hidden", (myipcache.a && wip.prefs["showmyasnumber"] && myip && wip.prefs["enablemyip"]) ? "false" : "true");

			panel_myasname.setAttribute("value", myascache.n);
			panel_myasname.setAttribute("style", "color:"+wip.prefs["showmyasnamecolor"]+";");
			panel_myasname.setAttribute("hidden", (myascache.n && wip.prefs["showmyasname"] && myip && wip.prefs["enablemyip"]) ? "false" : "true");
		} else {
			panel_mydatacenter.setAttribute("hidden", "true");
			panel_myasnumber.setAttribute("hidden", "true");
			panel_myasname.setAttribute("hidden", "true");
		}

		if (myipcache && myipcache.r) {
			panel_myrdns.setAttribute("value", myipcache.r);
			panel_myrdns.setAttribute("style", "color:"+wip.prefs["showmyrdnscolor"]+";");
			panel_myrdns.setAttribute("hidden", (myipcache.r && wip.prefs["showmyrdns"] && myip && wip.prefs["enablemyip"]) ? "false" : "true");
		}
		else
			panel_myrdns.setAttribute("hidden", "true");

		var ipcache = wip_storage.get("ip"+ip);
		var ascache = "";
		if (ipcache && ipcache.a)
			ascache = wip_storage.get("as"+ipcache.a);
		if (ascache) {
			var datacenter = ascache.d;
			if (datacenter.length>40)
				datacenter = datacenter.substring(0,40)+"…";
			panel_datacenter.setAttribute("value", datacenter);
			panel_datacenter.setAttribute("style", "color:"+wip.prefs["showdatacentercolor"]+";");
			panel_datacenter.setAttribute("hidden", (ascache.d && wip.prefs["showdatacenter"] && ip) ? "false" : "true");

			panel_asname.setAttribute("value", ascache.n);
			panel_asname.setAttribute("style", "color:"+wip.prefs["showasnamecolor"]+";");
			panel_asname.setAttribute("hidden", (ascache.n && wip.prefs["showasname"] && ip) ? "false" : "true");
		} else {
			panel_datacenter.setAttribute("hidden", "true");
			panel_asname.setAttribute("hidden", "true");
		}
		if (ipcache) {
			panel_asnumber.setAttribute("value", ipcache.a);
			panel_asnumber.setAttribute("style", "color:"+wip.prefs["showasnumbercolor"]+";");
			panel_asnumber.setAttribute("hidden", (ipcache.a && wip.prefs["showasnumber"] && ip) ? "false" : "true");

			panel_city.setAttribute("style", "color:"+wip.prefs["showcitycolor"]+";");
			panel_city.setAttribute("hidden", (ipcache.c && wip.prefs["showcity"] && ip) ? "false" : "true");

			panel_rdns.setAttribute("value", ipcache.r);
			panel_rdns.setAttribute("style", "color:"+wip.prefs["showrdnscolor"]+";");
			panel_rdns.setAttribute("hidden", (ipcache.r && wip.prefs["showrdns"] && ip) ? "false" : "true");
		} else {
			panel_asnumber.setAttribute("hidden", "true");
			panel_city.setAttribute("hidden", "true");
			panel_rdns.setAttribute("hidden", "true");
		}
	},

	setPanelAttributesAddressbar: function(ip)
	{
		var statusbar_flag = document.getElementById("worldip-statusbar-flag");
		var statusbar = statusbar_flag.getAttribute("src");
		var addressbar_flag = document.getElementById("worldip-addressbar-flag");
		addressbar_flag.setAttribute("src", statusbar);
		var panel_warning_red = document.getElementById("worldip-addressbar-warning-r");
		var panel_warning_yellow = document.getElementById("worldip-addressbar-warning-y");
		var panel_warning_green = document.getElementById("worldip-addressbar-warning-g");

		var hostActual = wip.getHostActual();
		var spoof = 0;
		if (hostActual)
			var spoof = wip_storage.get("spoof#"+wip.getHostActual());
		panel_warning_green.setAttribute("hidden", (spoof === 1 || spoof === 2) ? "false" : "true");
		panel_warning_yellow.setAttribute("hidden", (spoof === 3) ? "false" : "true");
		panel_warning_red.setAttribute("hidden", (spoof === 4) ? "false" : "true");
	},

	setPanelAttributes: function()
	{
		var ip = wip.getCurrentIP();
		var ip_v4 = (ip.indexOf(":")>=0) ? false : true;
		switch (wip.prefs["position"])
		{
			case "s": //statusbar
				var statusbar_panel = document.getElementById("worldip-statusbar");
				statusbar_panel.setAttribute("hidden", "false");
				var addressbar_panel = document.getElementById("worldip-addressbar");
				addressbar_panel.setAttribute("hidden", "true");
				wip.setPanelAttributesStatusbar(ip);
				break;
			case "a": //addressbar
				var statusbar_panel = document.getElementById("worldip-statusbar");
				statusbar_panel.setAttribute("hidden", "true");
				var addressbar_panel = document.getElementById("worldip-addressbar");
				addressbar_panel.setAttribute("hidden", "false");
				wip.setPanelAttributesAddressbar(ip);
		}
		var ipcache = wip_storage.get("ip"+ip);
		var ascache = "";
		if (ipcache && ipcache.a)
			ascache = wip_storage.get("as"+ipcache.a);
		else if (ipcache && !ipcache.l)
			ipcache.l = "--";
		// change variables in menu
		var ipmenu = document.getElementById("worldip-ip");
			ipmenu.setAttribute("label", ip ? (wip.worldip_ip_leer+ip) : wip.worldip_ip_leer);
			ipmenu.setAttribute("disabled", ip ? "false" : "true");
			if (ipcache && ipcache.l)
				ipmenu.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/"+ipcache.l+".png);");
			else
				ipmenu.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/--.png);");
		var whoisipmenu = document.getElementById("worldip-whoisip");
			whoisipmenu.setAttribute("label", ip ? ("Whois "+ip+"…") : "Whois");
			whoisipmenu.setAttribute("disabled", wip.isPublicNetz(ip) ? "false" : "true");
		var rdnsipmenu = document.getElementById("worldip-getrdns");
			rdnsipmenu.setAttribute("label", ip ? (wip.worldip_rdns_leer+ip) : "RDNS");
			if (ipcache && (ipcache.r && ipcache.r!=0))
				rdnsipmenu.setAttribute("disabled", "true");
			else
				rdnsipmenu.setAttribute("disabled", wip.isPublicNetz(ip) ? "false" : "true");
		var pingipmenu = document.getElementById("worldip-ping-ip-menu");
			pingipmenu.setAttribute("disabled", (wip.isPublicNetz(ip) && ip_v4) ? "false" : "true");
		var traceipmenu = document.getElementById("worldip-trace-ip-menu");
			traceipmenu.setAttribute("disabled", (wip.isPublicNetz(ip) && ip_v4) ? "false" : "true");
		var copyip = document.getElementById("worldip-copyip");
			copyip.setAttribute("disabled", ip ? "false" : "true");
		//var showlandinfo = document.getElementById("worldip-showlandinfo");
		//	showlandinfo.setAttribute("disabled", (wip.isPublicNetz(ip) && ip && ip_v4 && ipcache && ipcache.l && (ipcache.l!="--" || (ipcache.l=="--" && ipcache.c) ) && wip.currentAPI) ? "false" : "true");
		var reportbadland = document.getElementById("worldip-reportbadland");
			reportbadland.setAttribute("disabled", (wip.isPublicNetz(ip) && ip && ip_v4 && ipcache && ipcache.l && (ipcache.l!="--" || (ipcache.l=="--" && ipcache.c) ) && wip.currentAPI) ? "false" : "true");
		var changehostip = document.getElementById("worldip-changehostip");
			changehostip.setAttribute("disabled", (ip && (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"])) ? "false" : "true"); // wip.isPublicNetz(ip) && ip  - early, but for developers with server on localhost is unusable

		var checkall = document.getElementById("worldip-checkall");
			checkall.setAttribute("disabled", ip ? "false" : "true"); // wip.isPublicNetz(ip) && ip  - early, but for developers with server on localhost is unusable

		var myiplogmenu = document.getElementById("worldip-myiplog");
			if (wip.prefs["myiplogenable"])
				myiplogmenu.setAttribute("disabled", "false");
			else
				myiplogmenu.setAttribute("disabled", "true");
			
		//var getland = document.getElementById("worldip-getland");
		//var getipland = document.getElementById("worldip-getipland");
		//	if (wip.proxy_dns_enabled && !wip.prefs["proxy_dns_uselocal"]) {
		//		getipland.setAttribute("disabled", (!ip || (ipcache && ((!ipcache.l && !wip.prefs["enablelands"]) || (ipcache.l=="--" && !ipcache.c) ) && wip.isPublicNetz(ip) && wip.currentAPI && ip_v4)) ? "false" : "true");
		//		getland.setAttribute("disabled", "true");
		//	} else {
		//		getland.setAttribute("disabled", ( ipcache && ((!ipcache.l && !wip.prefs["enablelands"]) || (ipcache.l=="--" && !ipcache.c) ) && wip.isPublicNetz(ip) && wip.currentAPI && ip_v4) ? "false" : "true");
		//		getipland.setAttribute("disabled", "true");
		//	}

		var myip = wip.myip();
		var myipcache = wip_storage.get("ip"+myip);
		var worldipmyip = document.getElementById("worldip-myip");
			worldipmyip.setAttribute("label", (wip.prefs["enablemyip"]) ? (wip.worldip_myip_leer+myip) : wip.worldip_myip_leer);
			worldipmyip.setAttribute("disabled", (wip.prefs["enablemyip"]) ? "false" : "true");
			if (myip && myipcache && myipcache.l)
				worldipmyip.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/"+myipcache.l+".png);");
			else
				worldipmyip.setAttribute("style", "list-style-image: url(chrome://worldip/skin/flags"+wip.prefs["flagset"]+"/--.png);");
		var whoismyip = document.getElementById("worldip-whoismyip");
			whoismyip.setAttribute("label", (wip.prefs["enablemyip"]) ? ("Whois "+myip+"…") : "Whois");
			whoismyip.setAttribute("disabled", (wip.prefs["enablemyip"]) ? "false" : "true");
		var rdnsmyipmenu = document.getElementById("worldip-getmyrdns");
			rdnsmyipmenu.setAttribute("label", ip ? (wip.worldip_rdns_leer+myip) : "RDNS");
			if (myipcache && myipcache.r && myipcache.r!=0)
				rdnsmyipmenu.setAttribute("disabled", "true");
			else
				rdnsmyipmenu.setAttribute("disabled", wip.isPublicNetz(myip) ? "false" : "true");
		//var enablemyip = document.getElementById("worldip-enablemyip");
		//	enablemyip.setAttribute("checked", (wip.prefs["enablemyip"]) ? "true" : "false");
		var changemyip = document.getElementById("worldip-changemyip");
			changemyip.setAttribute("disabled", (wip.prefs["enablemyip"] && wip.currentAPI) ? "false" : "true");
		var copymyip = document.getElementById("worldip-copymyip");
			copymyip.setAttribute("disabled", (wip.prefs["enablemyip"] && wip.currentAPI) ? "false" : "true");
		var reportbadmyland = document.getElementById("worldip-reportbadmyland");
			reportbadmyland.setAttribute("disabled", (wip.prefs["enablemyip"] && wip.currentAPI) ? "false" : "true");
	},

	showMessage: function(message,clicked)
	{
		if (!message) return;
		try {
			var alertsService = WIPCc["@mozilla.org/alerts-service;1"].getService(WIPCi.nsIAlertsService);
			alertsService.showAlertNotification("chrome://worldip/skin/logo.png","WorldIP", message,clicked, "", null, "WorldipMessage");
		} catch(e) {}
	},

	hideTooltip: function()
	{
		document.getElementById( "worldip-statusbar" ).setAttribute( "tooltip", "" );
		document.getElementById( "worldip-addressbar" ).setAttribute( "tooltip", "" );
	},

	showTooltip: function()
	{
		document.getElementById( "worldip-statusbar" ).setAttribute( "tooltip", "worldip-tooltip" );
		document.getElementById( "worldip-addressbar" ).setAttribute( "tooltip", "worldip-tooltip" );
		document.getElementById( "worldip-tooltip" ).setAttribute( "noautohide", "true" );
	},

	freezeTooltip: function()
	{
		document.getElementById( "worldip-tooltip" ).setAttribute( "noautohide", "true" );
	},

	unfreezeTooltip: function()
	{
		document.getElementById( "worldip-tooltip" ).setAttribute( "noautohide", "false" );
	},

	renewTooltip: function()
	{
		var tooltip = document.getElementById("worldip-tooltipdata");
		while (tooltip.lastChild)
			tooltip.removeChild(tooltip.lastChild);
		if (!wip.currentUri)
			wip.currentUri = getBrowser().contentDocument.location;
		if (!wip.currentIP) {
			wip.currentIP = wip.getCurrentIP();
			if (!wip.currentIP) {
				var ips = 0;
				if (wip.currentUri && wip.currentUri.host)
					ips = wip_storage.get("dns#"+wip.currentUri.host);
				else
					ips = wip_storage.get("dns#"+wip.currentUri);
				if (ips && ips.length)
					wip.currentIP = ips[0];
			}
		}
		var ipcache = wip_storage.get("ip"+wip.currentIP);
		var countryCode = (ipcache && ipcache.l) ? ipcache.l : "";
		var tooltipdata = {};
		var stringbundle = document.getElementById("worldip-tooltiplabels");
		var PrivateType = wip.getPrivateType(wip.currentIP);
		if (wip.currentUri == "about:blank" || wip.currentUri == "about:newtab")
		{
			tooltipdata["hostname"] = "about:blank";
			tooltipdata["country"] = "localhost";
		}
		else if (wip.currentUri.host.indexOf(":")>0)
		{
			tooltipdata["hostname"] = wip.currentUri.host;
			tooltipdata["country"] = "V6";
		}
		else if (!countryCode)
		{
			try {
				tooltipdata["hostname"] = (wip.currentUri == "about:config") ? "about:config" : wip.currentUri.host;
			} catch(e) {
				tooltipdata["hostname"] = wip.currentUri;
			}
			if (wip.currentIP) {
				tooltipdata["hostip"] = wip.currentIP;
				var d = wip_storage.get("dns#"+wip.currentUri.host);
				tooltipdata["hostiprest"] = (d && d.length>1) ? d.slice(1) : "";
			}
			if (wip.currentAPI)
				tooltipdata["country"] = PrivateType ? PrivateType : "localhost";
			if (wip.proxy_dns_enabled)
				tooltipdata["country"] = "";
		}
		else
		{
			tooltipdata["hostname"] = wip.currentUri.host;
			tooltipdata["hostip"] = wip.currentIP;
			var d = wip_storage.get("dns#"+wip.currentUri.host);
			tooltipdata["hostiprest"] = (d && d.length>1) ? d.slice(1) : "";
			tooltipdata["country"] = countryCode;
			//tooltipdata["countrycode"] = countryCode;
		}
		if (ipcache)
		{
			if (ipcache.c) tooltipdata["city"] = ipcache.c; else tooltipdata["city"] = "";
			if (ipcache.r) tooltipdata["rdns"] = ipcache.r; else tooltipdata["rdns"] = "";
		}
		else
		{
			tooltipdata["city"] = "";
			tooltipdata["rdns"] = "";
		}
		var ascache = "";
		if (ipcache && ipcache.a)
			ascache = wip_storage.get("as"+ipcache.a);
		if (ascache)
		{
			if (ascache.d) {
				if (ascache.d.length>47) {
					var texts = ascache.d.match(/.{1,47}(\s+|$)/g);
					tooltipdata["datacenter"] = texts[0];
					for (t in texts)
						if (t!=0)
							tooltipdata["***"] = texts[t];
				}
				else
					tooltipdata["datacenter"] = ascache.d;
			}
			else tooltipdata["datacenter"] = "";
			tooltipdata["asnumber"] = ipcache.a;
			if (ascache.n)
				tooltipdata["asnumber"] += " ("+ascache.n+")";
				//tooltipdata["asname"] = ascache.n; else tooltipdata["asname"] = "";
		}
		else
		{
			tooltipdata["datacenter"] = "";
			tooltipdata["asnumber"] = "";
			//tooltipdata["asname"] = "";
		}
		if (tooltipdata["hostname"]) {
			var serverName = wip_storage.get("server#"+tooltipdata["hostname"]);
			if (serverName)
				tooltipdata["servername"] = serverName;
		}
		if (wip.currentIP && !PrivateType)
		{
			var ianastatus = wip.getIANAStatus(wip.currentIP);
			if (ianastatus[2]=="ALLOCATED")
				tooltipdata["ianastatus"] = ianastatus[0];
			else if (ianastatus[2]=="UNALLOCATED")
				tooltipdata["ianastatus"] = ianastatus[2];
			else
				tooltipdata["ianastatus"] = ianastatus[0]+" ("+ianastatus[2]+")";
		}

		var spoofType = wip_storage.get("spoof#"+tooltipdata["hostname"]);

		var isPublicNetz = wip.isPublicNetz(wip.currentIP);
		if ( (isPublicNetz || spoofType) && (wip.prefs["enablefulldns"] || wip.prefs["enablecheckextdns"]) && (!wip.proxy_dns_enabled || wip.prefs["proxy_dns_uselocal"]) )
			tooltipdata["extdns"] = true;

		var myip = wip.myip();
		if (wip.prefs["enablemyip"] && myip)
		{
			tooltipdata["myip"] = myip;
			var ipmyip = wip_storage.get("ip"+myip);
			if (ipmyip && ipmyip.l)
				tooltipdata["myland"] = wip_storage.get("ip"+myip).l;
			else if (!wip_storage.get("changemyiplock")) {
				wip_storage.set("changemyiplock", true, 5);
				wip.setTimer(function() { wip.changeMyIP(); }, 50, WIP_TYPE_ONE_SHOT);
			}
			//tooltipdata["mylandcode"] = wip_storage.get("ip"+myip).l;
		}

		var myipcache = wip_storage.get("ip"+myip);
		var myascache = "";
		if (myipcache && myipcache.a)
			myascache = wip_storage.get("as"+myipcache.a);
		if (myipcache && myipcache.r) tooltipdata["myrdns"] = myipcache.r; else tooltipdata["myrdns"] = "";
		if (myascache)
		{
			if (myascache.d) {
				if (myascache.d.length>47) {
						var texts = myascache.d.match(/.{1,47}(\s+|$)/g);
						tooltipdata["mydatacenter"] = texts[0];
						for (t in texts)
							if (t!=0)
								tooltipdata["***"] = texts[t];
				}
				else
					tooltipdata["mydatacenter"] = myascache.d;
			}
			else
				tooltipdata["mydatacenter"] = "";

			tooltipdata["myasnumber"] = myipcache.a;
			if (myascache.n)
				tooltipdata["myasnumber"] += " ("+myascache.n+")";
				//tooltipdata["myasname"] = myascache.n; else tooltipdata["myasname"] = "";
		}
		else
		{
			tooltipdata["mydatacenter"] = "";
			tooltipdata["myasnumber"] = "";
			//tooltipdata["myasname"] = "";
		}

		if (tooltipdata["hostname"] && spoofType) {
			var hbox = document.createElement("hbox");
			hbox.setAttribute("pack","start");
			hbox.setAttribute("align","start");
			var img = document.createElement("image");
			var spoofColor = ["","green","green","yellow","red"][spoofType];
			var spoofTextCode = wip_storage.get("spooftext#"+tooltipdata["hostname"]);
			try {
				if (spoofTextCode=="SpoofText_socks")
					var spoofText = stringbundle.getString(spoofTextCode);
				else
					var spoofText = stringbundle.getString("SpoofText_noip") + "\n" + stringbundle.getString(spoofTextCode);
				if (wip.currentIP && wip.isPublicNetz(wip.currentIP) && !tooltipdata["asnumber"] && spoofType>1)
					spoofText += "\n" + stringbundle.getString("SpoofText_asdisabled");
			} catch(e) {
				var spoofText = spoofTextCode;
			}
			if (spoofText) {
				var spoofTextArray = spoofText.split("\n").filter(function(x){return x});;
				img.setAttribute("src","chrome://worldip/skin/warning-"+spoofColor+".png");
				img.setAttribute("width","20px");
				img.setAttribute("height","16px");
				var vbox = document.createElement("vbox");
				for (s in spoofTextArray) {
					var value = document.createElement("label");
					var valuedata = spoofTextArray[s];
					value.appendChild(document.createTextNode(valuedata));
					value.style.whiteSpace = "pre";
					vbox.appendChild(value);
				}
				hbox.appendChild(img);
				hbox.appendChild(vbox);
				tooltip.appendChild(hbox);
			}
		}

		for (var element in tooltipdata)
		{
			if (element=="myip" || element=="hostname" || element=="extdns" )
				tooltip.appendChild(document.createElement("menuseparator"));
			if (element=="extdns" && !/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/.test(tooltipdata["hostname"]) ) {
				var menu = document.createElement("menu");
				menu.setAttribute("label", stringbundle.getString("externaldns"));
				menu.style.fontWeight = "bold";
				menu.onmouseover = function() {
					var anchor = document.getElementById("worldip-tooltip");
					var subpopup = document.getElementById("worldip-subtooltip");
					subpopup.openPopup(anchor,"start_before",-3,0,false);
				}
				menu.onmouseout = function() { var subpopup = document.getElementById("worldip-subtooltip"); subpopup.hidePopup(); }
				wip.renewSubTooltip(tooltipdata["hostname"], 0);
				tooltip.appendChild(menu);
			}
			if (tooltipdata[element]) {
				if (element=="hostiprest") {
					for ( i in tooltipdata[element] ) {
						var row = document.createElement("row");
						var label = document.createElement("label");
						var value = document.createElement("label");
						value.appendChild(document.createTextNode(tooltipdata[element][i]));
						value.style.whiteSpace = "pre";
						row.appendChild(label);
						row.appendChild(value);
						tooltip.appendChild(row);
					}
				} else if (typeof tooltipdata[element] == "string") {
					var row = document.createElement("row");
					if (element=="***")
						var labelstring = " ";
					else {
						try {
							var labelstring = stringbundle.getString(element) + ":";
						} catch (e) {
							var labelstring = element + ":";
						}
					}
					var label = document.createElement("label");
					label.appendChild(document.createTextNode(labelstring));
					label.style.fontWeight = "bold";
					var value = document.createElement("label");
					if (element == "country") {
						var valuedata = stringbundle.getString(tooltipdata[element]);
						if (countryCode)
							valuedata += " ("+countryCode+")";
					}
					else if (element == "myland" && tooltipdata["myland"])
						var valuedata = stringbundle.getString(tooltipdata[element])+" ("+tooltipdata[element]+")";
					else
						var valuedata = tooltipdata[element];
					value.appendChild(document.createTextNode(valuedata));
					value.style.whiteSpace = "pre";
					row.appendChild(label);
					row.appendChild(value);
					tooltip.appendChild(row);
				}
			}
		}

		document.getElementById( "worldip-statusbar" ).setAttribute( "tooltip", "worldip-tooltip" );
		document.getElementById( "worldip-tooltip" ).setAttribute( "noautohide", "true" );
	},

	renewSubTooltip: function(hostname, iter)
	{
		if (iter>4) return;
		iter += 1;
		var subtooltipdata = {};
		var extDNScurrent = wip_storage.get("extdns#"+hostname);
		var isPublicNetz = wip.isPublicNetz(wip.currentIP);
		var spoof = wip_storage.get("spoof#"+hostname);
		if (extDNScurrent && (isPublicNetz || spoof)) {
			var curTime = wip_storage.current_seconds();
			var extDNSs = [extDNScurrent,];
			var extDNSsDomains = [extDNScurrent["host"],];
			while (extDNScurrent && extDNScurrent["CNAME"] && extDNScurrent["CNAME"].length) {
				var cname = extDNScurrent["CNAME"][0]["primaryname"];
				extDNScurrent = wip_storage.get("extdns#"+cname);
				if (extDNScurrent) {
					if (extDNScurrent["host"]) {
						extDNSs.push(extDNScurrent);
						extDNSsDomains.push(extDNScurrent["host"]);
					} else if (extDNScurrent["error"]) {
						// todo???
					}
				} else {
					if (wip.prefs["enablefulldns"])
						wip.remoteDNS(cname, "ANY");
					else if (wip.prefs["enablecheckextdns"]) {
						wip.remoteDNS(cname, "A");
						//wip.remoteDNS(cname, "AAAA");
					}
					wip.setTimer(function() { wip.renewSubTooltip(hostname, iter); }, 300, WIP_TYPE_ONE_SHOT);
				}
			}
			var extDNScurrent = wip_storage.get("extdns#"+hostname);
			var ippattern = /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
			while (extDNScurrent && extDNScurrent["host"] && !ippattern.test(extDNScurrent["host"]) && (!extDNScurrent["NS"] || !extDNScurrent["NS"].length) ) {
				var tophost = extDNScurrent["host"].split(".").slice(1).join(".");
				extDNScurrent = wip_storage.get("extdns#"+tophost);
				if (extDNScurrent) {
					if (extDNScurrent["host"] && extDNSsDomains.indexOf(extDNScurrent["host"])<0)
						extDNSs.push(extDNScurrent);
					else if (extDNScurrent["error"]) {
						// todo???
					}
				} else if (tophost!=wip.tldService.getPublicSuffixFromHost(tophost)) {
					if (wip.prefs["enablefulldns"])
						wip.remoteDNS(tophost, "ANY");
					else if (wip.prefs["enablecheckextdns"]) {
						wip.remoteDNS(tophost, "A");
						//wip.remoteDNS(tophost, "AAAA");
						extDNScurrent = null;
					}
					wip.setTimer(function() { wip.renewSubTooltip(hostname, iter); }, 300, WIP_TYPE_ONE_SHOT);
				}
			}

			var tophost = wip.getTophost(hostname);
			if (tophost && extDNSsDomains.indexOf(tophost)<0) {
				extDNScurrent = wip_storage.get("extdns#"+tophost);
				if (extDNScurrent && extDNScurrent["host"]) {
					extDNSs.push(extDNScurrent);
				}
			}

			for (edns in extDNSs) {
				var extDNS = extDNSs[edns];
				var dnsdomain = extDNS["host"] ? extDNS["host"] : "";

				if (extDNS["error"]) {
					subtooltipdata["extdns"+hostname] = ["externaldns", extDNS[type], ""];
				} else {
					subtooltipdata["extdns"+dnsdomain] = ["", dnsdomain, "TTL"];
				}

				var reqTime = extDNS["reqtime"];
				var reqOld = curTime - reqTime;
				for (type in extDNS) {
					for (record in extDNS[type]) {
						if (type!="error")
							var entryTTL = (extDNS[type][record]["ttl"] > reqOld) ? extDNS[type][record]["ttl"]-reqOld : 0;
						switch (type) {
							case "A":
								subtooltipdata[type+record+dnsdomain] = [type, extDNS[type][record]["addr"], entryTTL];
								break;
							case "AAAA":
								subtooltipdata[type+record+dnsdomain] = [type, extDNS[type][record]["addr"], entryTTL];
								break;
							case "SPF":
								var text = extDNS[type][record]["spf"];
								if (text.length>55) {
									var texts = text.match(/.{1,55}(\s+|$)/g);
									subtooltipdata[type+record+dnsdomain+"0"] = [type, texts[0], entryTTL];
									for (t in texts)
										if (t!=0)
											subtooltipdata[type+record+dnsdomain+t] = ["", texts[t], ""];
								}
								else
									subtooltipdata[type+record+dnsdomain] = [type, text, entryTTL];
								break;
							case "TXT":
								var text = extDNS[type][record]["text"];
								if (text.length>55) {
									var texts = text.match(/.{1,55}(\s+|$)/g);
									subtooltipdata[type+record+dnsdomain+"0"] = [type, texts[0], entryTTL];
									for (t in texts)
										if (t!=0)
											subtooltipdata[type+record+dnsdomain+t] = ["", texts[t], ""];
								}
								else
									subtooltipdata[type+record+dnsdomain] = [type, text, entryTTL];
								break;
							case "MX":
								subtooltipdata[type+record+dnsdomain] = [type, extDNS[type][record]["mailserver"]+" ("+extDNS[type][record]["preference"]+")", entryTTL];
								break;
							case "PTR": // other case
								subtooltipdata[type+record+dnsdomain] = [type, extDNS[type][record]["domainname"], entryTTL];
								break;
							case "NAPTR": //ToDo
								break;
							case "RRSIG": //ToDo
								break;
							case "SRV": //ToDo
								break;
							case "NS":
								subtooltipdata[type+record+dnsdomain] = [type, extDNS[type][record]["nameserver"], entryTTL];
								break;
							case "SOA":
								subtooltipdata[type+record+dnsdomain+"mname"] = [type, "primary: "+extDNS[type][record]["mname"], entryTTL];
								subtooltipdata[type+record+dnsdomain+"rname"] = ["", "mailbox: "+extDNS[type][record]["rname"], ""];
								subtooltipdata[type+record+dnsdomain+"serial"] = ["", "serial: "+extDNS[type][record]["serial"], ""];
								subtooltipdata[type+record+dnsdomain+"refresh"] = ["", "refresh: "+extDNS[type][record]["refresh"], ""];
								subtooltipdata[type+record+dnsdomain+"retry"] = ["", "retry: "+extDNS[type][record]["retry"], ""];
								subtooltipdata[type+record+dnsdomain+"expire"] = ["", "expire: "+extDNS[type][record]["expire"], ""];
								subtooltipdata[type+record+dnsdomain+"minttl"] = ["", "minTTL: "+extDNS[type][record]["minttl"], ""];
								break;
							case "CNAME":
								subtooltipdata[type+dnsdomain] = [type, extDNS[type][record]["primaryname"], entryTTL];
								break;
						}
					}
				}
			}
		} else if (isPublicNetz || spoof) {
			subtooltipdata["extdns"] = ["externaldns", "ERROR", ""];
			if (wip.prefs["enablefulldns"])
				wip.remoteDNS(hostname, "ANY");
			else if (wip.prefs["enablecheckextdns"]) {
				wip.remoteDNS(hostname, "A");
				//wip.remoteDNS(hostname, "AAAA");
			}
			wip.setTimer(function() { wip.renewSubTooltip(hostname, iter); }, 300, WIP_TYPE_ONE_SHOT);
		}
		wip.generateSubPopup(subtooltipdata);
	},

	generateSubPopup: function(tooltipdata)
	{
		var subtooltip = document.getElementById("worldip-subtooltipdata");
		while (subtooltip.lastChild)
			subtooltip.removeChild(subtooltip.lastChild);

		for (var element in tooltipdata)
		{
			if (tooltipdata[element]) {
					var row = document.createElement("row");
					var label = document.createElement("label");
					if (tooltipdata[element][0]=="externaldns") {
						var stringbundle = document.getElementById("worldip-tooltiplabels");
						label.appendChild(document.createTextNode(stringbundle.getString("externaldns")));
					} else {
						label.appendChild(document.createTextNode(tooltipdata[element][0]));
					}
					var value = document.createElement("label");
					value.appendChild(document.createTextNode(tooltipdata[element][1]));
					value.style.whiteSpace = "pre";
					var value2 = document.createElement("label");
					value2.appendChild(document.createTextNode(tooltipdata[element][2]));
					if (tooltipdata[element][2]=="TTL" || tooltipdata[element][0]=="externaldns") {
						label.style.fontWeight = "bold";
						value.style.fontWeight = "bold";
						value2.style.fontWeight = "bold";
					} else if (tooltipdata[element][2]===0) {
						value2.style.color = "gray";
					}
					row.appendChild(label);
					row.appendChild(value);
					row.appendChild(value2);
					subtooltip.appendChild(row);
			}
		}
	},

	showAbout: function()
	{
		window.open("chrome://worldip/content/about.xul", "worldip", "centerscreen,chrome");
	},

	getMyASNumber: function() {
		var myASnumber = "";
		var myipcache = wip_storage.get("ip"+wip.myip());
		var myascache = "";
		if (myipcache && myipcache.a)
			myascache = wip_storage.get("as"+myipcache.a);
		if (myascache)
			myASnumber = myipcache.a;
		return myASnumber;
	},

	isGoogleServer: function(serverName) {
		if (!serverName)
			serverName = "";
		var serverNamePart = serverName.split("/")[0];
		var googleServers = ["gws","GFE","GSE","OFE","ucfe","pfe","mfe","SFE","cffe","codesite","NFE","mws","sffe","Google Frontend"];
		var isGoogleServer = googleServers.indexOf(serverNamePart)>=0;
		return isGoogleServer;
	},

	getSpoofLevel: function(ip, ips, ARecords, host) { // 0=nospoof 1=googlecache 2,3=cdn,bigcomp 4=spoof
		if (ip=="127.0.0.1") {// on my computer
			wip_storage.set("spooftext#"+host, "SpoofText_localhost", 3600);
			return 3;
		}
		if (!wip.isPublicNetz(ip)) {
			wip_storage.set("spooftext#"+host, "SpoofText_localnetz", 3600);
			return 4;
		}
		if (ip==wip.myip()) {
			wip_storage.set("spooftext#"+host, "SpoofText_myip", 3600);
			return 4;
		}
		var ipcache = wip_storage.get("ip"+ip);
		var myipcache = wip_storage.get("ip"+wip.myip());
		var asname = "";
		var asnumber = "";
		var land = (ipcache && ipcache.l) ? ipcache.l : "";
		var myland = (myipcache && myipcache.l) ? myipcache.l : "";
		var myASnumber = wip.getMyASNumber();
		if (ipcache && ipcache.a) {
			var asnumber = ipcache.a;
			var ascache = wip_storage.get("as"+ipcache.a);
			if (ascache && ascache.n) {
				var asnamefull = ascache.n.toLowerCase();
				var asname = asnamefull.split("-")[0]; // cut technical part, sometimes works
			}
		}

		var isSubnetz = false;
		for (i in ips) {
			if (!isSubnetz && wip.inSubnetwork(ips[i], ARecords))
				isSubnetz = true;
		}

		if (asnumber) {
			var tld = wip.tldService.getPublicSuffixFromHost(host);
			var basedomain = wip.tldService.getBaseDomainFromHost(host);
			var subDomainsInAS = {
				"google" : ["google."+tld, "youtube."+tld, "blogspot."+tld, "googlegroups."+tld, "googlecode."+tld, "googlelabs."+tld, "googleusercontent."+tld,  "googleapis."+tld, "appspot."+tld, "google-analytics."+tld, "googlesyndication."+tld, "doubleclick."+tld, "googleadservices."+tld, "googlepages."+tld, "googlesource."+tld,  "googlemail."+tld, "googleapps."+tld,  "googleratings."+tld, "withgoogle."+tld, "googledrive."+tld, "panoramio."+tld, "gogle."+tld, "googl."+tld, "whatbrowser."+tld, "ytimg."+tld,],
				"microsoft" : ["microsoft."+tld, "bing."+tld, "live."+tld, "msn."+tld,],
				"wikimedia" : ["wikipedia."+tld,],
			};
			subDomainsInAS["youtube"] = subDomainsInAS["google"];
			subDomainsInAS["cnnic-googlecn-ap"] = subDomainsInAS["google"];
			
			if (asname && (asname+"."+tld) == basedomain ) // Ex: GOOGLE.com == (accounts).google.com
				return 0;

			var googleTld = wip.isGoogleExtraTLD(tld);
			if ((asname=="google" || asname=="youtube") && googleTld )
				return 0;

			for (s in subDomainsInAS) {
				if ( subDomainsInAS[s].indexOf(basedomain)>=0 && (asname==s || asnamefull==s))
					return 0;
			}
			var serverName = wip_storage.get("server#"+host);
			var isGoogleServer = wip.isGoogleServer(serverName);
			var isGoogleDomain = subDomainsInAS["google"].indexOf(basedomain)>=0;
			if (isGoogleDomain || googleTld) {
				if (isSubnetz && (!serverName || isGoogleServer ) ) {
					wip_storage.set("spooftext#"+host, "SpoofText_subnetz", 3600);
					return 1;
				}
				if (asnumber && myASnumber==asnumber && (!serverName || isGoogleServer ) ) {
					wip_storage.set("spooftext#"+host, "SpoofText_ggcprov", 3600);
					return 1;
				}
				var gccAS = ["10292","10318","11664","11666","11758","1221","12332","12389","1241","12578","12615","1267","12683","12709","12713","12714","12741","12849","12956","12969","12975","1299","12997","13004","13184","13194","13285","13293","13489","13768","14615","15290","15589","15694","15735","16097","16586","17451","17494","17746","17806","17974","18101","18200","18678","18881","19092","19429","1955","19782","2018","20485","20500","20552","20960","210","21107","21211","21277","21351","21412","21453","21491","2200","22047","22742","22773","22799","23114","23655","23673","23700","23868","23930","23947","23951","24424","24523","24709","24835","24863","24955","25472","25490","2572","2603","2609","2764","27724","27948","28032","2828","28573","28580","28917","29422","29562","29614","31027","31042","31117","31133","31242","31585","3226","3239","3265","3292","3356","33922","33934","34123","34224","34547","34584","35002","35047","35254","35400","36947","36959","37228","3741","376","3786","3790","38082","38193","38506","38861","39107","39369","39392","39792","39912","41164","41313","41440","41920","4230","42610","42979","4323","43478","43515","43852","43975","44217","44775","45204","45430","45629","45727","45845","45899","45903","45904","4621","4648","4651","47232","47589","4761","4775","4780","4788","4800","48076","48176","4826","49624","49638","50542","50607","5078","5089","50923","52262","52376","5408","5483","5518","553","55329","5541","55818","5617","56246","57129","57193","57324","57344","577","6412","6471","6568","6677","6679","6700","6713","6730","6762","6855","6866","6871","6922","7018","7065","719","7303","7385","7418","7470","7545","7602","766","7738","7782","7795","8048","81","8218","8262","8308","8339","8346","8359","8400","8426","8447","8501","8544","855","8551","8612","8657","8708","8717","8764","8778","8781","8866","8897","8997","9002","9050","9116","9125","9198","9269","9286","9304","9329","9497","9498","9503","9587","9737","9822","9924",];
				if (gccAS.indexOf(asnumber)>=0 && (!serverName || isGoogleServer ) ) {
					wip_storage.set("spooftext#"+host, "SpoofText_ggcregio", 3600);
					if (myland==land)
						return 1;
					else
						return 3;
				} else if (isGoogleServer) {
					wip_storage.set("spooftext#"+host, "SpoofText_ggcunknown", 3600);
					return 4;
				}
			}
		}
		// Akamai & Co, but external DNS responses sometimes wrong because of many servers worldwide
		var cdn = {
			"akamai.net" : ["akamai",],
			"akamaiedge.net" : ["akamai",],
			"edgesuite.net" : ["akamai",],
			"v2cdn.net" : ["edgecast",],
			"edgecastcdn.net" : ["edgecast",],
			"cdngc.net" : ["cdnetworksus","cdnetworks",],
			"gccdn.net" : ["cdnetworksus","cdnetworks",],
			"netdna-cdn.com" : [],
			"cdn77.net" : [],
			"ccgslb.com" : [], //chinacache.com
			"c3cdn.net" : [], //chinacache.com
			"hwcdn.net" : ["highwinds","highwinds2","highwinds3","highwinds5",],
			"llnwd.net" : ["llnw",],
			"cloudapp.net" : ["microsoft",],
			"ecedexis.net" : [],
			"nyucd.net" : [], //coral
			"cotcdn.net" : ["coten",],
			"incapdns.net" : ["incapsula",],
			"cloudflare.net" : ["cloudflarenet",],
			"cachefly.net" : ["cachenetworks",],
			"cloudfront.net" : ["amazon",],
			"opendns.com" : ["opendns",],
			"amazonaws.com" : ["amazon",],
		}
		var cdnns = {
			"cloudflare.com" : ["cloudflarenet",],
			"opendns.com" : ["opendns",],
		}
		var bigcomp = {
			"bankone.com" : ["7743",],
			"bnpparibas.net" : ["25215",],
		}
		var bigcompns = {
			"chase.com" : ["7743",],
			"bankone.com" : ["7743",],
			"jpmorganchase.com" : ["7743",],
		}
		var cnames = wip.getCNAMErecords(host);
		for (num in cnames) {
			var cnametop = wip.getTophost(cnames[num]);
			if (cnametop && cdn[cnametop] && asnumber) {
				if (asnumber && myASnumber==asnumber) {
					wip_storage.set("spooftext#"+host, "SpoofText_cdnprov", 3600);
					return 1;
				}
				if (cdn[cnametop].length && cdn[cnametop].indexOf(asname)>=0) {
					wip_storage.set("spooftext#"+host, "SpoofText_cdnself", 3600);
					return 1;
				}
				if (!isSubnetz) {
					wip_storage.set("spooftext#"+host, "SpoofText_cdn", 3600);
					return 3;
				}
			}
			if (cnametop && bigcomp[cnametop] && asnumber) {
				if (bigcomp[cnametop].length && bigcomp[cnametop].indexOf(asnumber)>=0) {
					wip_storage.set("spooftext#"+host, "SpoofText_bigcomp", 3600);
					return 2;
				}
			}
		}
		var extDNS = wip_storage.get("extdns#"+host);
		if (extDNS && extDNS["NS"] && extDNS["NS"].length)  {
			var nsservtop = wip.getTophost(extDNS["NS"][0]["nameserver"]);
			if (nsservtop && cdnns[nsservtop] && cdnns[nsservtop].length && cdnns[nsservtop].indexOf(asname)>=0) {
				wip_storage.set("spooftext#"+host, "SpoofText_cdnself", 3600);
				return 1;
			}
			if (nsservtop && bigcompns[nsservtop] && bigcompns[nsservtop].length && asnumber && bigcompns[nsservtop].indexOf(asnumber)>=0) {
				wip_storage.set("spooftext#"+host, "SpoofText_bigcomp", 3600);
				return 2;
			}
		}

		if (isSubnetz) {
			wip_storage.set("spooftext#"+host, "SpoofText_subnetz", 3600);
			return 2;
		}

		if (ARecords.length && asname=="amazon") {
			for (i in ARecords) {
				var amazonIP = wip_storage.get("ip"+ARecords[i]);
				if (amazonIP && amazonIP.a) {
					var amazonAS = wip_storage.get("as"+amazonIP.a);
					if (amazonAS && amazonAS.n && (amazonAS.n.toLowerCase().split("-")[0]=="amazon") ) {
						wip_storage.set("spooftext#"+host, "SpoofText_cdn", 3600);
						return 2;
					}
				}
			}
		}

		if (asnumber && myASnumber==asnumber) {
			wip_storage.set("spooftext#"+host, "SpoofText_spoofprov", 3600);
		}
		else
			wip_storage.set("spooftext#"+host, "SpoofText_spoof", 3600);
		return 4;
	},

	getAllIP4: function(host) {
		var d = wip_storage.get("dns#"+host);
		var ip4 = [];
		if (d && d.length) {
			for (n in d) {
				if ( d[n].indexOf(':')<0 && /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(d[n]) ) {
					ip4.push(d[n]);
				}
			}
		}
		return ip4;
	},

	inSubnetwork: function(ip, IPs) {
		var ipOctets = ip.split(".");
		try {
			var ip3 = ipOctets[0]+"."+ipOctets[1]+"."+ipOctets[2];
			for (n in IPs) {
				var ipOctets = IPs[n].split(".");
				if (ip3 == (ipOctets[0]+"."+ipOctets[1]+"."+ipOctets[2]) )
					return true;
			}
		} catch(e) {}
		return false;
	},

	arrayCompareHit: function(ips1, ips2) {
		for (i in ips1) {
			var pos = ips2.indexOf(ips1[i]);
			if (pos>=0)
				return ips2[pos];
		}
		return null;
	},

	isGoogleExtraTLD: function(tld) {
		var googleDomainsTLD = ["blogspot", "googleapis", "googlecode", "appspot"];
		if (tld && googleDomainsTLD.indexOf(tld.split(".")[0])>=0)
			return true;
		return false;
	},

	getTLD: function(host) {
		var tld = wip.tldService.getPublicSuffixFromHost(host);
		return tld;
	},

	getTophost: function(host) {
		if (!host) return null;
		if (/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(host)) return null;
		var tophost = wip.tldService.getBaseDomainFromHost(host);
		var tld = wip.getTLD(host);
		var ausnahmeTLD = ["cloudfront.net","amazonaws.com"];
		if (tld && ausnahmeTLD.indexOf(tld)>=0)
			return tld;
		var isGoogleTLD = wip.isGoogleExtraTLD(tld);
		if (tld!=tophost && tophost!="com" && tophost!="net" && tophost!="org")
			return tophost;
		else if (isGoogleTLD)
			return tld;
		return null;
	},

	compareArecords: function(host, stage)
	{
		var ARecords = wip.getArecords(host);
		var ip = wip.ipResolve(host, true);
		var ips = wip.getAllIP4(host);

		if (ip && !wip.arrayCompareHit(ips, ARecords)) {
			var spoof = wip.getSpoofLevel(ip, ips, ARecords, host);
		} else if (ip) {
			var spoof = 0;
		} else {
			wip.ipResolve(host);
			return;
		}
		if (stage==1)
			wip_storage.delete("spoofpopuplock#"+host);
		if (spoof<3 || (stage>=4 && ARecords.length))
			wip_storage.set("spoof#"+host, spoof, 24*3600);
		if (spoof==0 || (stage<4 && spoof>2))
			wip_storage.delete("spooftext#"+host);
		if (spoof && spoof>1) {
			var t = 0;
			var tophost = wip.getTophost(host);
			if (!wip.isPublicNetz(ip) || ip==wip.myip() || ip=="127.0.0.1")
				stage==4;
			if (stage>=4)
				wip.loaderDisable(host);
			else
				wip.loaderEnable(host);
			var extDNS = wip_storage.get("extdns#"+host);
			if (stage==1 || !ARecords.length) {
				var t = 30;
				if (!wip.prefs["enablefulldns"] && !(extDNS && extDNS["NS"] && extDNS["NS"].length) && tophost ) {
					wip.setTimer(function() { wip.remoteDNS( tophost , "NS", 0, true, wip.dnstcp); }, t, WIP_TYPE_ONE_SHOT);
					t += 150;
				}
				wip.setTimer(function() { wip.remoteDNS(host, "A", 0, true, wip.dnstcp); }, t, WIP_TYPE_ONE_SHOT);
				t += 150;
				wip.setTimer(function() { wip.remoteDNS(host, "ANY", 0, true, wip.dnstcp); }, t, WIP_TYPE_ONE_SHOT);
			}
			if (stage==2) {
				var serv = "";
				if (extDNS && extDNS["NS"] && extDNS["NS"].length) {
					var serv = extDNS["NS"][0]["nameserver"];
				} else {
					var serv = wip.getNSServFromTophost(host);
				}
				var nstop = wip.getTophost(serv);
				if (nstop && nstop.split(".")[0].split("-")[0]=="awsdns") {
					var ipcache = wip_storage.get("ip"+ip);
					if (ipcache && ipcache.a) {
						var ascache = wip_storage.get("as"+ipcache.a);
						if (ascache && ascache.n) {
							var asname = ascache.n.toLowerCase().split("-")[0];
							if (asname=="amazon" && ARecords.length)
								wip.setTimer(function() { wip.getData(ARecords[0],"",false); }, 30, WIP_TYPE_ONE_SHOT);
						}
					}
				}
				var t = 30;
				if (serv) {
					wip.setTimer(function() { wip.remoteDNS(host, "A", 0, true, "udp", serv); }, t, WIP_TYPE_ONE_SHOT);
					t += 200;
					wip.setTimer(function() { wip.remoteDNS(host, "ANY", 0, true, "udp", serv); }, t, WIP_TYPE_ONE_SHOT);
					t += 150;
				}
				if (tophost && tophost!=host)
					wip.setTimer(function() { wip.remoteDNS(tophost, "ANY", 0, true, wip.dnstcp); }, t, WIP_TYPE_ONE_SHOT);
			}
			if (stage==3 && spoof==4) {
				var t = 30;
				var isGoogleServer = wip.isGoogleServer( wip_storage.get("server#"+host) );
				if (isGoogleServer) {
					wip.setTimer(function() { wip.remoteDNS(host, "A", 0, true, wip.dnstcp, "8.8.8.8"); }, t, WIP_TYPE_ONE_SHOT);
					t += 150;
					wip.setTimer(function() { wip.remoteDNS(host, "ANY", 0, true, wip.dnstcp, "8.8.8.8"); }, t, WIP_TYPE_ONE_SHOT);
					t += 150;
				}
				if (extDNS && extDNS["CNAME"] && extDNS["CNAME"].length) {
						var cname = extDNS["CNAME"][0]["primaryname"];
						wip.setTimer(function() { wip.remoteDNS(cname, "ANY", 0, true, wip.dnstcp); }, t, WIP_TYPE_ONE_SHOT);
						t += 300;
						wip.setTimer(function() { wip.remoteDNS(cname, "A", 0, true, wip.dnstcp); }, t, WIP_TYPE_ONE_SHOT);
				}
				wip_storage.set("comparearecslock#"+host, true, 1);
				wip.dnsAsyncRequest(host, true, false, true);
			}
			if (stage<4)
				wip.setTimer(function() { wip.compareArecords(host, (stage+1)); }, t+500, WIP_TYPE_ONE_SHOT);
			else if (stage==4 && spoof==4) {
				for (var i=1; i<12; i++) {
					wip.setTimer(function f(x){wip.compareArecords(host, (stage+x));}.bind(null, i), (i*i*200+400) , WIP_TYPE_ONE_SHOT);
				}
				wip.setTimer(function() { wip.showSpoofPopup(host, ip); }, 3800, WIP_TYPE_ONE_SHOT);
			}
			else if (stage==6 && spoof==4 && ARecords.length) {
				wip.setTimer(function() { wip.requestAPI(wip.currentAPI, "spoofcheck/?s=4&h="+host+"&ip="+ip, wip.remoteSpoofCheckCallback, {"host":host,"ip":ip}); }, 50, WIP_TYPE_ONE_SHOT);
			}
		} else
			wip.loaderDisable(host);
		wip.showInfoActual();
	},

	showSpoofPopup: function(host, ip) {
		var spoof = wip_storage.get("spoof#"+host);
		var actualhost = wip.getHostActual();
		if (spoof==4 && wip.prefs["enablespoofpopup"] && !wip_storage.get("spoofpopuplock#"+host) && (actualhost==host)) { //|| actualhost=="api.wipmania.com" - think about geodns
			var ipcache = wip_storage.get("ip"+ip);
			if (ipcache && ipcache.a) {
				wip_storage.set("spoofpopuplock#"+host, spoof, 3600);
				wip.showMessage("DNS-Spoofing: "+host, false);
			}
		}
	},

	remoteSpoofCheckCallback: function(doc, args)
	{
		switch (doc) // ToDo more
		{
			case "Y": // yes
				wip_storage.set("spoof#"+args["host"], 4, 24*3600);
				wip.setTimer(function() { wip.showSpoofPopup(args["host"], args["ip"]); }, 30, WIP_TYPE_ONE_SHOT);
				break;
			case "N": // no
				wip_storage.set("spoof#"+args["host"], 0, 24*3600);
				wip_storage.set("spoofpopuplock#"+args["host"], 1, 5);
				wip_storage.delete("spooftext#"+args["host"]);
				break;
			case "U": // unknown
				break;
			default: // some error
				break;
		}
	},

	extractARecordsFromDNS: function(extDNScurrent) {
		var extARecords = [];
		if (extDNScurrent) {
			if (extDNScurrent["A"])
				for (a in extDNScurrent["A"])
					if (extDNScurrent["A"][a]["addr"])
						extARecords.push(extDNScurrent["A"][a]["addr"]);
			if (extDNScurrent["AAAA"])
				for (a in extDNScurrent["AAAA"])
					if (extDNScurrent["AAAA"][a]["addr"])
						extARecords.push(extDNScurrent["AAAA"][a]["addr"]);
		}
		return extARecords;
	},

	arrayUnique: function(array) {
		var a = array.concat();
		for(var i=0; i<a.length; ++i) {
			for(var j=i+1; j<a.length; ++j) {
				if(a[i] === a[j])
					a.splice(j--, 1);
			}
		}
		return a;
	},

	getNSServFromTophost: function(host) {
		var extDNSfirst = wip_storage.get("extdns#"+host);
		if (!extDNSfirst)	return "";

		var extDNScurrent = extDNSfirst;
		if (extDNScurrent["NS"] && extDNScurrent["NS"].length)
			return extDNScurrent["NS"][0]["nameserver"];

		var extDNScurrent = extDNSfirst;
		while (extDNScurrent && extDNScurrent["CNAME"] && extDNScurrent["CNAME"].length) {
			extDNScurrent = wip_storage.get("extdns#"+extDNScurrent["CNAME"][0]["primaryname"]);
			if (extDNScurrent && extDNScurrent["NS"] && extDNScurrent["NS"].length)
				return extDNScurrent["NS"][0]["nameserver"];
		}

		var extDNScurrent = extDNSfirst;
		while (extDNScurrent && extDNScurrent["host"]) {
			var tophost = extDNScurrent["host"].split(".").slice(1).join(".");
			extDNScurrent = wip_storage.get("extdns#"+tophost);
			if (extDNScurrent && extDNScurrent["NS"] && extDNScurrent["NS"].length)
				return extDNScurrent["NS"][0]["nameserver"];
		}

		var extDNScurrent = wip_storage.get("extdns#"+wip.tldService.getBaseDomainFromHost(host));
		if (extDNScurrent && extDNScurrent["NS"] && extDNScurrent["NS"].length)
			return extDNScurrent["NS"][0]["nameserver"];
		else
			var extDNScurrent = wip_storage.get("extdns#"+host);
		return "";
	},

	getArecords: function(host) {
		var extARecords = [];
		var extDNScurrent = wip_storage.get("extdns#"+host);
		if (!extDNScurrent)	return [];
		extARecords = wip.extractARecordsFromDNS(extDNScurrent);

		while (extDNScurrent && extDNScurrent["CNAME"] && extDNScurrent["CNAME"].length) {
			var primaryname = extDNScurrent["CNAME"][0]["primaryname"];
			extDNScurrent = wip_storage.get("extdns#"+primaryname);
			if (extDNScurrent)
				extARecords = extARecords.concat( wip.extractARecordsFromDNS(extDNScurrent) );
		}

		var extDNScurrent = wip_storage.get("extdns#"+host);
		while (extDNScurrent && extDNScurrent["host"]) {
			var tophost = extDNScurrent["host"].split(".").slice(1).join(".");
			extDNScurrent = wip_storage.get("extdns#"+tophost);
			if (extDNScurrent)
				extARecords = extARecords.concat( wip.extractARecordsFromDNS(extDNScurrent) );
		}

		tophost = wip.getTophost(host);
		extDNScurrent = wip_storage.get("extdns#"+tophost);
		if (extDNScurrent)
			extARecords = extARecords.concat( wip.extractARecordsFromDNS(extDNScurrent) );

		extARecords.sort();
		extARecords = wip.arrayUnique(extARecords);
		return extARecords;
	},

	getCNAMErecords: function(host) {
		var extDNScurrent = wip_storage.get("extdns#"+host);
		if (!extDNScurrent || !extDNScurrent["CNAME"]) return [];
		var extCNAMERecords = [];
		while (extDNScurrent && extDNScurrent["CNAME"] && extDNScurrent["CNAME"].length) {
			var primaryname = extDNScurrent["CNAME"][0]["primaryname"];
			extCNAMERecords.push(primaryname);
			extDNScurrent = wip_storage.get("extdns#"+primaryname);
		}
		extCNAMERecords.sort();
		return extCNAMERecords;
	},

	MyIPLogTabl: function(myiplogpanel)
	{
		if (!wip.prefs["myiplogenable"])
			return;
		var myiplog = wip_storage.get("myiplog");
		if (!myiplog)
			return "";
		var all_data = "";
		for (l in myiplog) {
			var ipcache = wip_storage.get("ip"+myiplog[l].ip);
			if (ipcache == null || !ipcache.l)
				ipcache = {l:"", a:""};
			if (ipcache.a)
				var ascache = wip_storage.get("as"+ipcache.a);
			else
				var ascache = {n:"",d:""};
			var data = myiplog[l].ip + "	"+ ipcache.l + "	" + myiplog[l].time + "	" + " AS" + ipcache.a + "	" + ascache.d;
			if (data.length>100)
				data = data.substring(0,100)+"…";
			all_data += data+"\n";
		}
		myiplogpanel.setAttribute( "value", all_data );
	},
	
	showMyIPLog: function()
	{
		var w = window.open("chrome://worldip/content/myip.xul", "worldiplog", "centerscreen,chrome,resizable=yes,scrollbars=yes");
		wip.setTimer(function() {
			var myiplogpanel = w.document.getElementById("myiplogpanel-text");
			wip.MyIPLogTabl(myiplogpanel);
		}, 50, WIP_TYPE_ONE_SHOT);
	},

	showOptions: function()
	{
		var optwin = WIPCc['@mozilla.org/appshell/window-mediator;1'].getService(WIPCi.nsIWindowMediator).getMostRecentWindow("WorldIP:Options");
		if (optwin)
			optwin.focus();
		else
			openDialog("chrome://worldip/content/options.xul", "", "centerscreen,chrome,titlebar,toolbar");
	},

	goHome: function(url)
	{
		var browser = document.getElementById("content");
		var tab = browser.addTab("http://www.wipmania.com/"+url);
		browser.selectedTab = tab;
	},

	getCurrentIP: function()
	{
		var ipe = document.getElementById("worldip-statusbar-ip");
		var ipline = ipe.getAttribute("value");
		return ipline.split(" ")[0];
	},

	copyIP: function(ip)
	{
		const copytoclip = WIPCc["@mozilla.org/widget/clipboardhelper;1"]
			.getService(WIPCi.nsIClipboardHelper);
		copytoclip.copyString(ip);
	},


	runProgram: function(path, arg) // Used only for run of ping and traceroute commands. See following functions runLocalPing and runLocalTrace
	{
		var args = arg ? arg.split(/\s+/) : [];
		var nsIFile = WIPCc["@mozilla.org/file/local;1"].getService(WIPCi.nsIFile);
		var nsIProcess = WIPCc["@mozilla.org/process/util;1"].createInstance(WIPCi.nsIProcess);
		nsIFile.initWithPath(path);
		nsIProcess.init(nsIFile);
		nsIProcess.run(false, args, args.length);
	},

	runLocalPing: function(ip)
	{
		var pattern = /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/;
		if (!pattern.test(ip)) return;
		if (WIP_OS=="WINNT")
			wip.runProgram("C:\\WINDOWS\\system32\\cmd.exe","/C start cmd.exe /K ping -n 5 "+ip);
		else //Linux, Darwin, etc
			wip.runProgram("/usr/bin/xterm","-hold -e ping -c 5 "+ip);
	},

	runLocalTrace: function(ip)
	{
		var pattern = /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/;
		if (!pattern.test(ip)) return;
		if (WIP_OS=="WINNT")
			wip.runProgram("C:\\WINDOWS\\system32\\cmd.exe","/C start cmd.exe /K tracert "+ip);
		else
			wip.runProgram("/usr/bin/xterm","-hold -e traceroute "+ip);
	},

	showLandInfo: function(ip)
	{
		var land = (wip_storage.get("ip"+ip) && wip_storage.get("ip"+ip).l) ? wip_storage.get("ip"+ip).l : "";
		if (ip && land && wip.isPublicNetz(ip))
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab("http://www.wipmania.com/map/"+land+"/");
			browser.selectedTab = tab;
		}
	},

	reportBadLand: function(ip)
	{
		var land = wip_storage.get("ip"+ip) ? wip_storage.get("ip"+ip).l : "";
		if (ip && land && wip.isPublicNetz(ip) )
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab("http://www.wipmania.com/base/ip/?ip="+ip+"&ff");
			browser.selectedTab = tab;
		}
	},

	goHomeTrace: function(ip,city)
	{
		if (ip && wip.isPublicNetz(ip))
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab("http://www.wipmania.com/tracefrom/"+ip+"/?c="+city+"&ff=");
			browser.selectedTab = tab;
		}
	},

	goHomePing: function(ip,city)
	{
		if (ip && wip.isPublicNetz(ip))
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab("http://www.wipmania.com/pingfrom/"+ip+"/?c="+city+"&ff=");
			browser.selectedTab = tab;
		}
	},

	goHomeWhois: function(ip)
	{
		if (ip && wip.isPublicNetz(ip))
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab("http://www.wipmania.com/whois/"+ip+"/?ff");
			browser.selectedTab = tab;
		}
	},

	goProvider: function(provurl)
	{
		if (provurl)
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab(provurl);
			browser.selectedTab = tab;
		}
	},

	enableMyIP: function(from_menu)
	{
		if (wip.prefs["enablemyip"])
		{
			if (from_menu)
				wip.preferences.setBoolPref("enablemyip",false);
			wip.prefs["enablemyip"]= wip.preferences.getBoolPref("enablemyip");
			wip_storage.set("myip", "");
			wip_storage.set("myiplast", "");
			wip.setPanelAttributes();
		}
		else
		{
			if (from_menu)
				wip.preferences.setBoolPref("enablemyip",true);
			wip.prefs["enablemyip"]= wip.preferences.getBoolPref("enablemyip");
			wip.changeMyIP();
		}
		if (!wip.prefs["enablemyip"])
			try { wip.MyIPTimer.cancel(); } catch(e) {}
		document.getElementById( "worldip-statusbar" ).setAttribute( "tooltip", "worldip-tooltip" );
	},

	changeMyIP: function()
	{
		if (wip.prefs["enablemyip"]) {
			if (wip.prefs["updatemyipinterval"] > 0)
			{
				try { wip.MyIPTimer.cancel(); } catch(e) {}
				var event = { notify: function(timer) { wip.getMyIP(); } };
				try { wip.MyIPTimer.initWithCallback(event, (wip.prefs["updatemyipinterval"]*1000), WIP_TYPE_REPEATING_SLACK); } catch(e) {}
			}
			wip.getMyIP();
		}
		else
		{
			try { wip.MyIPTimer.cancel(); } catch(e) {}
			wip_storage.set("myip", "");
			wip_storage.set("myiplast", "");
			wip.setPanelAttributes();
		}
		document.getElementById( "worldip-statusbar" ).setAttribute( "tooltip", "worldip-tooltip" );
	},

	changeHostIP: function()
	{
		if (wip.currentUri && wip.currentUri.host)
		{
			var host = wip.currentUri.host;
			wip_storage.delete("dns#"+host);
			if (host.indexOf(':')<0 && !(/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(host)) ) {
				wip.dnsAsyncRequest(host, true, false, false);
			}
		}
	},

	getRDNScallback: function(doc, ip)
	{
		var pattern = /^[-0-9a-z\.]+$/i;
		if (doc && !pattern.test(doc)) doc = 0;
		if (!doc && doc==="") doc = "not found";
		if (ip) {
			var ipcache = wip_storage.get("ip"+ip);
			if (ipcache) {
				ipcache.r = doc;
				wip_storage.set("ip"+ip, ipcache, 24*3600);
			}
			else
				wip_storage.set("ip"+ip, {"l":"--","a":"","r":doc,"c":""});
		}
		wip.setPanelAttributes();
	},

	getRDNS: function(ip)
	{
		if (!ip || !wip.isPublicNetz(ip))
			return false;
		var ipcache = wip_storage.get("ip"+ip);
		if (ipcache && ipcache.r && ipcache.r!=0)
			return false;

		var ipcachedns = wip_storage.get("extdns#"+ip);
		if (ipcachedns && ipcachedns["PTR"] && ipcachedns["PTR"].length)
			wip.getRDNScallback(ipcachedns["PTR"][0]["domainname"], ip);
		else
			wip.remoteDNS(ip, "PTR");
	},

	showASonHE: function(ip)
	{
		var ipcache = wip_storage.get("ip"+ip);
		if (!ipcache || !ipcache.a)
			return;
		if (ip && wip.isPublicNetz(ip))
		{
			var browser = document.getElementById("content");
			var tab = browser.addTab("http://bgp.he.net/AS"+ipcache.a);
			browser.selectedTab = tab;
		}
	},

	onBarClick: function(e)
	{
		if (e && (e.button == 0) && (!e.ctrlKey && !e.shiftKey))
		{
			var action = wip.prefs["clickaction"];
			switch(action)
			{
				case "no":
					break;
				case "getall":
					wip.getAndCheckAll();
					break;
				case "copy":
					wip.copyIP(wip.getCurrentIP());
					break;
				case "land":
					wip.showLandInfo(wip.getCurrentIP());
					break;
				case "rdns":
					wip.getRDNS(wip.getCurrentIP());
					break;
				case "myrdns":
					wip.getRDNS(wip.myip());
					break;
				case "newip":
					wip.changeMyIP();
					break;
				case "wrongland":
					wip.reportBadLand(wip.getCurrentIP());
					break;
				case "getipland":
					wip.ipRemoteResolve();
					break;
				case "getland":
					wip.getDisabledLand();
					break;
				case "asonhe":
					wip.showASonHE(wip.getCurrentIP());
					break;
			}
		}
	},

	// DNS Part

	randomDNSServer: function() {
		var extdnsservers = wip.prefs["extdnsservers"].split("\n").map(function(v) { return (/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(v) ? v : ""); }).filter(function(x){return x});
		var rnd = Math.floor(Math.random()*extdnsservers.length);
		return extdnsservers[rnd];
	},

	remoteDNS: function(hostraw, type, iter, force, proto, serv) {
		if (!serv)
			serv = "";
		if (!proto)
			proto = "udp";
		if ( !hostraw || (serv && wip_storage.get("extdnslock#"+type+"#"+serv+"#"+hostraw+"#"+proto)) || hostraw=="about:newtab" || hostraw=="about:blank") return;
		if (iter && iter>3) return; // try 3 times if errors
		else if (!iter)
			iter = 0;
		iter += 1;
		var host = hostraw.split(':')[0];
		var ippattern = /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
		if (type!="PTR" && ippattern.test(host) ) return;
		var cacheDNS = wip_storage.get("extdns#"+host);
		if (!force && cacheDNS && cacheDNS[type] && cacheDNS[type].length) return;
		if (!serv)
			var server = wip.randomDNSServer();
		else
			var server = serv;
		if (server) {
			if ( wip_storage.get("extdnslock#"+type+"#"+server+"#"+host+"#"+proto) ) return;
			var respData = wip_storage.get("extdnslock#"+type+"#"+server+"#"+host+"#");
			if (respData) {
				wip.remoteDNScallback(host, type, respData, 0, "", server);
			}
			else {
				wip_storage.set("extdnslock#"+type+"#"+server+"#"+host+"#"+proto, true, 5);
				wip.setTimer(function() { wip.remoteDNSRequest(host, type, server, proto, iter); }, 30, WIP_TYPE_ONE_SHOT);
			}
		}
	},

	remoteDNSerrback: function(status, host, type, iter, protocol, server) {
		wip_storage.delete("extdnslock#"+type+"#"+server+"#"+host+"#"+protocol);
		var timeout = 5;
		switch(status) {
			case WIP_DNS_CODE_ERROR_FORMAT:
				timeout = 3600;
				var error = "ERROR_FORMAT";
				break;
			case WIP_DNS_CODE_ERROR_SERVFAIL:
				timeout = 3600;
				var error = "ERROR_SERVER_FAILURE";
				break;
			case WIP_DNS_CODE_ERROR_NAME:
				timeout = 3600;
				var error = "ERROR_NAME";
				break;
			case WIP_DNS_CODE_ERROR_NOT_IMPLEMENTED:
				timeout = 3600;
				var error = "ERROR_NOT_IMPLEMENTED";
				break;
			case WIP_DNS_CODE_ERROR_REFUSED:
				timeout = 10;
				var error = "ERROR_DNS_CONNECTION_REFUSED";
				break;
			case WIP_ERROR_NET_RESET:
				var error = "ERROR_NET_RESET";
				break;
			case WIP_ERROR_NET_TIMEOUT:
				var error = "ERROR_NET_TIMEOUT";
				break;
			case WIP_ERROR_CONNECTION_REFUSED:
				var error = "ERROR_CONNECTION_REFUSED";
				break;
			default:
				var error="ERROR_UNKNOWN";
		}
		if (iter>3 || ["ERROR_NOT_IMPLEMENTED", "ERROR_NAME", "ERROR_FORMAT" ].indexOf(error)>=0 ) {
			var extdns = wip_storage.get("extdns#"+host);
			if (!extdns)
				wip_storage.set("extdns#"+host, {"error":error, "host":host}, timeout);
			if (type=="PTR") {
				wip.getRDNScallback( ((error == "ERROR_NAME") ? "" : 0) , host);
			}
		} else {
			if (status==WIP_ERROR_NET_TIMEOUT && protocol=="tcp" && host=="api.wipmania.com") {
				wip.dnstcp = "udp"; //shade...
				wip.remoteDNS(host, type, iter, true, "udp", server);
			}
			else {
				wip.remoteDNS(host, type, iter, true, protocol);
			}
		}
	},

	arrayAUnique: function(array, n) {
		var a = array.concat();
		for(var i=0; i<a.length; ++i) {
			for(var j=i+1; j<a.length; ++j) {
				if(a[i][n] === a[j][n])
					a.splice(j--, 1);
			}
		}
		return a;
	},

	remoteDNScallback: function(host, type, records, minTTL, protocol, server) {
		if (minTTL<3600) minTTL=3600;
		var cacheDNS = wip_storage.get("extdns#"+host);
		if (!cacheDNS)
			cacheDNS = {};
		else if (cacheDNS["error"])
			delete cacheDNS["error"];
		for (r in records) {
				if (records[r].length || r=="host" || r=="reqtime") {
					if (cacheDNS[r] && (r=="A" || r=="AAAA" || r=="NS") ) {
						var n = (r=="NS") ? "nameserver" : "addr";
						cacheDNS[r] = wip.arrayAUnique(cacheDNS[r].concat(records[r]), n);
					}
					else
						cacheDNS[r] = records[r];
				}
		}
		wip_storage.set("extdns#"+host, cacheDNS, minTTL);
		if (protocol)
			wip_storage.set("extdnslock#"+type+"#"+server+"#"+host+"#", records, 6);
		var ippattern = /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
		if ( (!cacheDNS["NS"] || !cacheDNS["NS"].length) && type!="PTR" && type!="A" && type!="AAAA" && !ippattern.test(host) ) {
			var mainhost = wip.getTophost(host);
			if (mainhost && mainhost!=host)
				wip.setTimer(function() { wip.remoteDNS(mainhost, "ANY"); }, 50, WIP_TYPE_ONE_SHOT);
		}
		if (cacheDNS["CNAME"] && cacheDNS["CNAME"].length && type=="ANY") {
			var cname = cacheDNS["CNAME"][0]["primaryname"];
			if (cname && cname!=host)
				wip.setTimer(function() { wip.remoteDNS(cname, "ANY"); }, 100, WIP_TYPE_ONE_SHOT);
		}
		if (type=="PTR" && records["PTR"]) {
			wip.getRDNScallback(records["PTR"][0]["domainname"], host);
		}
	},

	decodeUTF8: function(utf) {
		var i=0;
		var txt = "";
		while (i<utf.length) {
			var u1 = utf.charCodeAt(i);
			if (u1<128) {
				txt += String.fromCharCode(u1);
				i++;
			}
			else if((u1>191) && (u1<224)) {
				var u2 = utf.charCodeAt(i+1);
				txt += String.fromCharCode(((u1&31)<<6)|(u2&63));
				i+=2;
			}
			else {
				var u2 = utf.charCodeAt(i+1);
				var u3 = utf.charCodeAt(i+2);
				txt += String.fromCharCode(((u1&15)<<12)|((u2&63)<<6)|(u3&63));
				i+=3;
			}
		}
		return txt;
	},

	hex2chars: function(s) {
		if (!s)
			return "";
		s = s.replace(/^(0x)?/g,"");
		s = s.replace(/[^a-fA-F0-9]/g,"");
		s = s.split("");
		var result = "";
		for (var i=0; i<s.length; i+=2) {
			result += String.fromCharCode(parseInt(s[i]+''+s[i+1],16));
		}
		return result;
	},

	chars2hex: function(s) {
		var n = null;
		var result = "";
		for (var i=0; i<s.length; i++) {
			n = s.charCodeAt(i);
			result += WIP_HEX[(n >> 4) & 0xf] + WIP_HEX[n & 0xf];
		}
		return result;
	},

	decompressDNSName: function(pointer, dnsBody, text) {
		if (parseInt("0x"+dnsBody[pointer]+dnsBody[pointer+1], 16) > 191) {// pointer
			var newPointer = parseInt(("0x"+dnsBody[pointer]+dnsBody[pointer+1]) & 0x3f, 16)*256+parseInt(dnsBody[pointer+2]+''+dnsBody[pointer+3],16);
			return wip.decompressDNSName(newPointer*2, dnsBody, text);
		}
		else if (dnsBody[pointer]+dnsBody[pointer+1] == "00") { // end 
			return text;
		}
		else { // count
			var count = parseInt(dnsBody[pointer]+''+dnsBody[pointer+1],16);
			if (text) text += "2e";
			try {
				return wip.decompressDNSName(pointer + 2 + count*2, dnsBody, text+dnsBody.substring(pointer+2,pointer+2+count*2));
			} catch(e) {return null;}
		}
	},

	parseDNSRecord: function(type, typeBody, dnsBody, pointer) {
		var RRdata = {};
		switch (type) {
			case "A":
				var AArray = typeBody.split(/(.{2})/).filter(function(x){return x});
				AArray = AArray.map(function(v) { return parseInt(v,16); });
				RRdata["addr"] = AArray.join(".");
				break;
			case "AAAA":
				var AArray = typeBody.split(/(.{4})/).filter(function(n){return n});
				AArray = AArray.map(function(v) { return v.replace(/^0+/, ''); });
				RRdata["addr"] = AArray.join(":");
				if ( (RRdata["addr"].match(/::/g) || []).length > 0 && (RRdata["addr"].match(/:::/g) || []).length == 0 )
					RRdata["addr"] = RRdata["addr"].replace("::",":0:");
				while ((RRdata["addr"].match(/:::/g) || []).length) RRdata["addr"] = RRdata["addr"].replace(/:::/g,"::");
				break;
			case "SPF":
				RRdata["spf"] = wip.hex2chars(typeBody.substring(2));
				break;
			case "TXT":
				RRdata["text"] = wip.decodeUTF8(wip.hex2chars(typeBody.substring(2)));
				break;
			case "MX":
				RRdata["preference"] = parseInt("0x"+typeBody.substr(0,4), 16);
				RRdata["mailserver"] = wip.hex2chars(wip.decompressDNSName(pointer+4, dnsBody, ""));
				break;
			case "PTR":
				RRdata["domainname"] = wip.hex2chars(wip.decompressDNSName(pointer, dnsBody, ""));
				break;
			case "NAPTR": //ToDo
				break;
			case "NS":
				RRdata["nameserver"] = wip.hex2chars(wip.decompressDNSName(pointer, dnsBody, ""));
				break;
			case "SOA":
				RRdata["mname"] = wip.hex2chars(wip.decompressDNSName(pointer, dnsBody, "")); // primary name server
				var mnameLen = RRdata["mname"].length;
				var mnameBody = typeBody.substr(0,mnameLen*2+4);
				var mnameBodyNoRedirectArray = mnameBody.split(/(.{2})/).filter(function(x){return x}).map(function(v) { return v.replace(/c[012]/,""); });
				var mnameBodyNoRedirect = "";
				var i = 0;
				while (mnameBodyNoRedirectArray[i]) {
					mnameBodyNoRedirect += mnameBodyNoRedirectArray[i];
					i +=1 ;
				}
				if (mnameBodyNoRedirect.length < mnameBody.length)
					mnameBodyNoRedirect += "0000";
				var rnamePointer = pointer + mnameBodyNoRedirect.length;
				RRdata["rname"] = wip.hex2chars(wip.decompressDNSName(rnamePointer, dnsBody, "")); // responsible authority's mailbox
				RRdata["serial"] = parseInt("0x"+typeBody.substr(typeBody.length-40,8));
				RRdata["refresh"] = parseInt("0x"+typeBody.substr(typeBody.length-32,8));
				RRdata["retry"] = parseInt("0x"+typeBody.substr(typeBody.length-24,8));
				RRdata["expire"] = parseInt("0x"+typeBody.substr(typeBody.length-16,8));
				RRdata["minttl"] = parseInt("0x"+typeBody.substr(typeBody.length-8,8));
				break;
			case "CNAME":
				RRdata["primaryname"] = wip.hex2chars(wip.decompressDNSName(pointer, dnsBody, ""));
				break;
			case "SRV":
				break;
			default:;
		}
		return RRdata;
	},

	remoteDNSRequest: function(domain, type, server, protocol, iter, isPingTimeout) {
		if (!domain) return;
		var tcp = (protocol.toLowerCase()=="tcp") ? 2 : 0;
		var tcpShift = tcp ? 4 : 0;
		var tcpLength = tcp ? wip.chars2hex(wip.hex2chars((domain.length+18).toString(16))) : "";
		if (tcpLength.length==2) tcpLength = "00" + tcpLength;

		var QueryID = WIP_HEX[Math.floor(Math.random() * 16)] + WIP_HEX[Math.floor(Math.random() * 16)] + WIP_HEX[Math.floor(Math.random() * 16)] + WIP_HEX[Math.floor(Math.random() * 16)];
		var QueryType = WIP_DNS_QTYPE[type];
		var domainArray = domain.split(".");
		var domainLen = domain.length;
		if (type=="PTR") {
			domainArray.reverse();
			domainArray = domainArray.concat(["in-addr","arpa"]);
			domainLen += 13;
		}
		var query = "";
		for (var i=0; i<domainArray.length; i++) {
			query += (domainArray[i].length<16 ? "0":"") + domainArray[i].length.toString(16) + wip.chars2hex(domainArray[i]);
		}
		var udp = tcpLength + QueryID+"01000001000000000000" + query + "0000"+QueryType+"0001";
		var SocketTransportService = WIPCc["@mozilla.org/network/socket-transport-service;1"].getService(WIPCi.nsISocketTransportService);
		try {
			if (tcp)
				var Transport = SocketTransportService.createTransport(0,0,server,53,null);
			else
				var Transport = SocketTransportService.createTransport(["udp",],1,server,53,null);
		} catch(e){ return; }
		var TransportIn = Transport.openInputStream(0,0,0);
		var TransportOut = Transport.openOutputStream(0,0,0);
		Transport.setTimeout(1, 5); //bug in FF - for TCP doesn't work, fast always 30sec
		var udpChars = wip.hex2chars(udp);
		if (isPingTimeout) {
			if (!iter) {
				TransportIn.close();TransportOut.close();
				return;
			}
			var startTime = new Date().getTime();
		}
		TransportOut.write(udpChars, domainLen+18+tcp);

		var dnsDataListener =
		{
			onStartRequest: function(request, context){},
			onStopRequest: function(request, context, status)
			{
				TransportIn.close();TransportOut.close();
				if (status)
					wip.remoteDNSerrback(status, domain, type, iter, protocol, server);
			},
			onDataAvailable: function(request, context, inputStream, offset, count) {
				var rawread = InputStream.readBytes(TransportIn.available());
				TransportIn.close();TransportOut.close();
				if (rawread.length) {
					var hexread = wip.chars2hex(rawread).substring(tcpShift);
					if (hexread.substring(0,4) == QueryID && hexread.substring(24,24+query.length) == query) {
						var flag1 = parseInt("0x"+hexread.substring(4,6));
						var flag2 = parseInt("0x"+hexread.substring(6,8));
						var flagResponse = flag1 & 0x80 ? true : false;
						var flagAuthoritative = flag1 & 0x04 ? true : false;
						var flagTruncated = flag1 & 0x02 ? true : false;
						var flagRecursionDesired = flag1 & 0x01 ? true : false;
						var flagRecursionAvailable = flag2 & 0x80 ? true : false;
						var flagAnswerAuthenticated = flag2 & 0x20 ? true : false;
						var replyCode = flag2 & 0x0f;
						var AnswerCount = parseInt("0x"+hexread.substring(12,16),16); // how much NS
						var Authoritycount = parseInt("0x"+hexread.substring(16,20),16);

						if (replyCode==0 && flagTruncated && type!="PTR") {
							wip.setTimer(function() { wip.remoteDNSRequest(domain, type, server, "tcp", iter); }, 30, WIP_TYPE_ONE_SHOT);
						}
						else if (replyCode==0 && (AnswerCount+Authoritycount)>0) {
							var response = hexread.substring(34+query.length);
							var s = response.split('');
							var respData = {"A":[],"AAAA":[],"TXT":[],"MX":[],"PTR":[],"NAPTR":[],"NS":[],"SOA":[],"CNAME":[],"SRV":[],"SPF":[],"RRSIG":[],"host":domain,"reqtime":wip_storage.current_seconds()};
							var minTTL = 0;
							for (var i=0; i<s.length; i+=2) {
									var RRdomain = wip.hex2chars(wip.decompressDNSName(i+34+query.length, hexread, ""));
									var RRtype = WIP_DNS_RQTYPE[s[i+6]+s[i+7]];
									var RRttl = parseInt(s[i+12]+''+s[i+13]+s[i+14]+s[i+15]+s[i+16]+s[i+17]+s[i+18]+s[i+19], 16);
									var RRlen = parseInt('0x'+s[i+20]+s[i+21]+s[i+22]+s[i+23], 16)*2;
									if (RRtype) {
										var RRdata = wip.parseDNSRecord(WIP_DNS_RQTYPE[s[i+6]+s[i+7]], response.substr(i+24, RRlen), hexread, 34+query.length+i+24);
										RRdata["ttl"] = RRttl;
										RRdata["name"] = RRdomain;
										if (!minTTL || minTTL > RRttl)
											minTTL = RRttl;
										respData[RRtype].push(RRdata);
									}
									i = i + 22 + RRlen;
							}
							wip.remoteDNScallback(domain, type, respData, minTTL, protocol, server);
						}
						else {
							wip.remoteDNSerrback(replyCode, domain, type, iter, protocol, server);
						}
					}
				}
			},
		};

		var InputStream = WIPCc["@mozilla.org/binaryinputstream;1"].createInstance(WIPCi.nsIBinaryInputStream);
		InputStream.setInputStream(TransportIn);
		if (isPingTimeout) {
			while (!TransportIn.available() && (new Date().getTime() - startTime)<isPingTimeout) 1;
			var dnsDistance = new Date().getTime() - startTime;
			var rawread = InputStream.readBytes(TransportIn.available());
			TransportIn.close();TransportOut.close();
			wip.setTimer(function() { wip.remoteDNSRequest(domain, type, server, protocol, iter-1, isPingTimeout); }, 1000, WIP_TYPE_ONE_SHOT);
			wip.dnsPingLog(domain, type, server, dnsDistance, rawread.length, iter);
		} else {
			var InputStreamPump = WIPCc["@mozilla.org/network/input-stream-pump;1"].createInstance(WIPCi.nsIInputStreamPump);
			InputStreamPump.init(TransportIn, -1, -1, 0, 0, true);
			InputStreamPump.asyncRead(dnsDataListener,null);
		}
	},

	dnsCustom: function()
	{
		var w = window.open("chrome://worldip/content/dnscustom.xul", "worldipdnstools", "centerscreen,chrome,resizable=yes,scrollbars=yes");
	},

	dnsPingLog: function(domain, type, server, time, isOk, iter)
	{
		var w = window.open("", "worldipdnstools");
		var worldipdnslog = w.document.getElementById("worldipDNSLog");
		var dataLabel = w.document.getElementById("worldipDNSCustomData");
		var data = dataLabel.getAttribute("value");
		var oldval = worldipdnslog.getAttribute("value");
		if (!oldval)
			oldval = 'DNS-PING: '+server+' -t '+type+' "'+domain+'"\n---------------------------\n';
		var newdata = isOk ? "time="+time+" ms" : "timeout";
		var newdataTime = isOk ? time : "0";
		dataLabel.setAttribute( "value", data+newdataTime+" ");
		if (iter==1) {
			var data = dataLabel.getAttribute("value").split(" ").filter(function(x){return x}).map(function(v){return parseInt(v);});
			data.sort();
			var dataok = data.filter(function(x){return x});
			var dataoklen = dataok.length;
			var sum = dataok.reduce(function(pv, cv) { return pv + cv; }, 0);
			var sum2 = dataok.reduce(function(pv, cv) { return pv + cv*cv; }, 0);
			var datamin = dataoklen ? dataok[0] : "-";
			var datamax = dataoklen ? dataok[dataoklen-1] : "-";
			var dataavg = dataoklen ? (sum/dataoklen).toFixed(2) : "-";
			var datamdev = dataoklen ? Math.sqrt( sum2/dataoklen - (sum*sum)/(dataoklen*dataoklen) ).toFixed(2) : "-";
			newdata += "\n\n--- dns ping statistics ---\n5 packets transmitted, "+dataoklen+" received, "+((data.length-dataoklen)*20)+"% packet loss\nrtt min/avg/max/mdev = "+datamin+"/"+dataavg+"/"+datamax+"/"+datamdev+" ms";
		}
		worldipdnslog.setAttribute( "value", oldval+"req="+(6-iter)+" "+newdata+"\n");
	},

	dnsCustomGetValues: function() {
		var w = window.open("", "worldipdnstools");
		var dnsserver = w.document.getElementById("worldipDNSCustomServer").value;
		var dnstarget = w.document.getElementById("worldipDNSCustomTarget").value;
		w.document.getElementById("worldipDNSLog").setAttribute( "value", "");
		w.document.getElementById("worldipDNSCustomData").setAttribute( "value", "");
		var type = "A";
		if (dnstarget) {
			if ( /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(dnstarget) ) type = "PTR";
		} else
			dnstarget = "www.google.com";
		if (dnsserver) {
			if (!/\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/.test(dnsserver)) {
				try {
					wip.dns.resolve(dnsserver, 0);
				} catch(e) {
					w.document.getElementById("worldipDNSLog").setAttribute( "value", "NS_ERROR_UNKNOWN_HOST");
					return;
				}
			}
		} else
			dnsserver = wip.randomDNSServer();
		if (!dnsserver)
			dnsserver = "8.8.8.8";
		return [dnstarget, type, dnsserver];
	},

	dnsCustomRequest: function(iterCheck)
	{
		var v = wip.dnsCustomGetValues();
		if (v[1]!="PTR")
			v[1] = "ANY";
		var dnstarget = v[0];
		var type = v[1];
		var dnsserver =  v[2];
		if (!iterCheck) {
			wip_storage.delete("extdns#"+v[0]);
			wip.setTimer(function() { wip.remoteDNSRequest(dnstarget, type, dnsserver, "udp", 4); }, 50, WIP_TYPE_ONE_SHOT);
			wip.setTimer(function() { wip.dnsCustomRequest(1); }, 500, WIP_TYPE_ONE_SHOT);
			return;
		}
		var w = window.open("", "worldipdnstools");
		var worldipdnslog = w.document.getElementById("worldipDNSLog");
		if (iterCheck>6) {
			worldipdnslog.setAttribute( "value", "DNS_TIMEOUT");
			return;
		}
		var extDNS = wip_storage.get("extdns#"+dnstarget);
		if (!extDNS) {
			wip.setTimer(function() { wip.dnsCustomRequest(iterCheck+1); }, 500, WIP_TYPE_ONE_SHOT);
			return;
		}
		if (extDNS["error"]) {
			worldipdnslog.setAttribute( "value", extDNS["error"]);
			return;
		}
		var val = 'DNS-REQUEST: '+dnsserver+' -t '+type+' "'+dnstarget+'"\n---------------------------\n';
		for (type in extDNS) {
			var entryTTL = extDNS[type].length ? extDNS[type][0]["ttl"] : 0;
			var entryTTLtext = (type.length>3) ? " (TTL: "+entryTTL+")" : " (TTL: "+entryTTL+")";
			if (type!="host" && type!="reqtime")
				val += type+entryTTLtext+"\n";
			for (record in extDNS[type]) {
				if (type=="host") continue;
				switch (type) {
							case "A":
								val += "		"+extDNS[type][record]["addr"]+"\n";
								break;
							case "AAAA":
								val += "		"+extDNS[type][record]["addr"]+"\n";
								break;
							case "SPF":
								val += "		"+extDNS[type][record]["spf"]+"\n";
								break;
							case "TXT":
								val += "		"+extDNS[type][record]["text"]+"\n";
								break;
							case "MX":
								val += "		"+extDNS[type][record]["mailserver"]+" ("+extDNS[type][record]["preference"]+")\n";
								break;
							case "PTR":
								val += "		"+extDNS[type][record]["domainname"]+"\n";
								break;
							case "NS":
								val += "		"+extDNS[type][record]["nameserver"]+"\n";
								break;
							case "SOA":
								val += "		primary:"+extDNS[type][record]["mname"]+"\n";
								val += "		mailbox:"+extDNS[type][record]["rname"]+"\n";
								val += "		serial:"+extDNS[type][record]["serial"]+"\n";
								val += "		refresh:"+extDNS[type][record]["refresh"]+"\n";
								val += "		retry:"+extDNS[type][record]["retry"]+"\n";
								val += "		expire:"+extDNS[type][record]["expire"]+"\n";
								val += "		minttl:"+extDNS[type][record]["minttl"]+"\n";
								break;
							case "CNAME":
								val += "		"+extDNS[type][record]["primaryname"]+"\n";
								break;
				}
			}
		}
		worldipdnslog.setAttribute( "value", val);
	},

	dnsPing: function()
	{
		var v = wip.dnsCustomGetValues();
		wip.setTimer(function() { wip.remoteDNSRequest(v[0], v[1], v[2], "udp", 5, 2000); }, 50, WIP_TYPE_ONE_SHOT);
	},

}
